##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Basic Site content type registry
"""

import os
import re
from six.moves import urllib

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from App.special_dtml import DTMLFile
from OFS.SimpleItem import SimpleItem
from Persistence import PersistentMapping
from zope.component import getUtility
from zope.interface import implementer
from ZPublisher.mapply import mapply

from Products.CMFCore.interfaces import IContentTypeRegistry
from Products.CMFCore.interfaces import IContentTypeRegistryPredicate
from Products.CMFCore.interfaces import ITypesTool
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import _dtmldir
from Products.CMFCore.utils import registerToolInterface


@implementer(IContentTypeRegistryPredicate)
class MajorMinorPredicate(SimpleItem):

    """
        Predicate matching on 'major/minor' content types.
        Empty major or minor implies wildcard (all match).
    """

    major = minor = None
    PREDICATE_TYPE = 'major_minor'

    security = ClassSecurityInfo()

    def __init__(self, id):
        self.id = id

    @security.protected(ManagePortal)
    def getMajorType(self):
        """ Get major content types.
        """
        if self.major is None:
            return 'None'
        return ' '.join(self.major)

    @security.protected(ManagePortal)
    def getMinorType(self):
        """ Get minor content types.
        """
        if self.minor is None:
            return 'None'
        return ' '.join(self.minor)

    @security.protected(ManagePortal)
    def edit(self, major, minor, COMMA_SPLIT=re.compile(r'[, ]')):

        if major == 'None':
            major = None
        if type(major) is type(''):
            major = [_f for _f in COMMA_SPLIT.split(major) if _f]

        if minor == 'None':
            minor = None
        if type(minor) is type(''):
            minor = [_f for _f in COMMA_SPLIT.split(minor) if _f]

        self.major = major
        self.minor = minor

    #
    #   ContentTypeRegistryPredicate interface
    #
    security.declareObjectPublic()
    def __call__(self, name, typ, body):
        """
            Return true if the rule matches, else false.
        """
        if self.major is None:
            return 0

        if self.minor is None:
            return 0

        typ = typ or '/'
        if not '/' in typ:
            typ = typ + '/'
        major, minor = typ.split('/', 1)

        if self.major and not major in self.major:
            return 0

        if self.minor and not minor in self.minor:
            return 0

        return 1

    @security.protected(ManagePortal)
    def getTypeLabel(self):
        """
            Return a human-readable label for the predicate type.
        """
        return self.PREDICATE_TYPE

    security.declareProtected(ManagePortal, 'predicateWidget')
    predicateWidget = DTMLFile('majorMinorWidget', _dtmldir)

InitializeClass(MajorMinorPredicate)


@implementer(IContentTypeRegistryPredicate)
class ExtensionPredicate(SimpleItem):

    """
        Predicate matching on filename extensions.
    """

    extensions = None
    PREDICATE_TYPE = 'extension'

    security = ClassSecurityInfo()

    def __init__(self, id):
        self.id = id

    @security.protected(ManagePortal)
    def getExtensions(self):
        """ Get filename extensions.
        """
        if self.extensions is None:
            return 'None'
        return ' '.join(self.extensions)

    @security.protected(ManagePortal)
    def edit(self, extensions, COMMA_SPLIT=re.compile(r'[, ]')):

        if extensions == 'None':
            extensions = None
        if type(extensions) is type(''):
            extensions = [_f for _f in COMMA_SPLIT.split(extensions) if _f]

        self.extensions = extensions

    #
    #   ContentTypeRegistryPredicate interface
    #
    security.declareObjectPublic()
    def __call__(self, name, typ, body):
        """
            Return true if the rule matches, else false.
        """
        if self.extensions is None:
            return 0

        _base, ext = os.path.splitext(name)
        if ext and ext[0] == '.':
            ext = ext[1:]

        return ext in self.extensions

    @security.protected(ManagePortal)
    def getTypeLabel(self):
        """
            Return a human-readable label for the predicate type.
        """
        return self.PREDICATE_TYPE

    security.declareProtected(ManagePortal, 'predicateWidget')
    predicateWidget = DTMLFile('extensionWidget', _dtmldir)

InitializeClass(ExtensionPredicate)


@implementer(IContentTypeRegistryPredicate)
class MimeTypeRegexPredicate(SimpleItem):

    """
        Predicate matching only on 'typ', using regex matching for
        string patterns (other objects conforming to 'match' can
        also be passed).
    """

    pattern = None
    PREDICATE_TYPE = 'mimetype_regex'

    security = ClassSecurityInfo()

    def __init__(self, id):
        self.id = id

    @security.protected(ManagePortal)
    def getPatternStr(self):
        if self.pattern is None:
            return 'None'
        return self.pattern.pattern

    @security.protected(ManagePortal)
    def edit(self, pattern):
        if pattern == 'None':
            pattern = None
        if type(pattern) is type(''):
            pattern = re.compile(pattern)
        self.pattern = pattern

    #
    #   ContentTypeRegistryPredicate interface
    #
    security.declareObjectPublic()
    def __call__(self, name, typ, body):
        """
            Return true if the rule matches, else false.
        """
        if self.pattern is None:
            return 0

        return self.pattern.match(typ)

    @security.protected(ManagePortal)
    def getTypeLabel(self):
        """
            Return a human-readable label for the predicate type.
        """
        return self.PREDICATE_TYPE

    security.declareProtected(ManagePortal, 'predicateWidget')
    predicateWidget = DTMLFile('patternWidget', _dtmldir)

InitializeClass(MimeTypeRegexPredicate)


@implementer(IContentTypeRegistryPredicate)
class NameRegexPredicate(SimpleItem):

    """
        Predicate matching only on 'name', using regex matching
        for string patterns (other objects conforming to 'match'
        and 'pattern' can also be passed).
    """

    pattern = None
    PREDICATE_TYPE = 'name_regex'

    security = ClassSecurityInfo()

    def __init__(self, id):
        self.id = id

    @security.protected(ManagePortal)
    def getPatternStr(self):
        """
            Return a string representation of our pattern.
        """
        if self.pattern is None:
            return 'None'
        return self.pattern.pattern

    @security.protected(ManagePortal)
    def edit(self, pattern):
        if pattern == 'None':
            pattern = None
        if type(pattern) is type(''):
            pattern = re.compile(pattern)
        self.pattern = pattern

    #
    #   ContentTypeRegistryPredicate interface
    #
    security.declareObjectPublic()
    def __call__(self, name, typ, body):
        """
            Return true if the rule matches, else false.
        """
        if self.pattern is None:
            return 0

        return self.pattern.match(name)

    @security.protected(ManagePortal)
    def getTypeLabel(self):
        """
            Return a human-readable label for the predicate type.
        """
        return self.PREDICATE_TYPE

    security.declareProtected(ManagePortal, 'predicateWidget')
    predicateWidget = DTMLFile('patternWidget', _dtmldir)

InitializeClass(NameRegexPredicate)


_predicate_types = []

def registerPredicateType(typeID, klass):
    """
        Add a new predicate type.
    """
    _predicate_types.append((typeID, klass))

for klass in (MajorMinorPredicate, ExtensionPredicate, MimeTypeRegexPredicate,
              NameRegexPredicate):
    registerPredicateType(klass.PREDICATE_TYPE, klass)


@implementer(IContentTypeRegistry)
class ContentTypeRegistry(SimpleItem):

    """
        Registry for rules which map PUT args to a CMF Type Object.
    """

    meta_type = 'Content Type Registry'
    id = 'content_type_registry'

    manage_options = (
        ({'label': 'Predicates', 'action': 'manage_predicates'},
         {'label': 'Test', 'action': 'manage_testRegistry'}) +
        SimpleItem.manage_options)

    security = ClassSecurityInfo()

    def __init__(self):
        self.predicate_ids = ()
        self.predicates = PersistentMapping()

    #
    #   ZMI
    #
    security.declarePublic('listPredicateTypes')
    def listPredicateTypes(self):
        """
        """
        return [x[0] for x in _predicate_types]

    security.declareProtected(ManagePortal, 'manage_predicates')
    manage_predicates = DTMLFile('registryPredList', _dtmldir)

    @security.protected(ManagePortal)
    def doAddPredicate(self, predicate_id, predicate_type, REQUEST):
        """
        """
        self.addPredicate(predicate_id, predicate_type)
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_predicates'
                                     + '?manage_tabs_message=Predicate+added.')

    @security.protected(ManagePortal)
    def doUpdatePredicate(self, predicate_id, predicate, typeObjectName,
                          REQUEST):
        """
        """
        self.updatePredicate(predicate_id, predicate, typeObjectName)
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                   + '/manage_predicates'
                                   + '?manage_tabs_message=Predicate+updated.')

    @security.protected(ManagePortal)
    def doMovePredicateUp(self, predicate_id, REQUEST):
        """
        """
        predicate_ids = list(self.predicate_ids)
        ndx = predicate_ids.index(predicate_id)
        if ndx == 0:
            msg = "Predicate+already+first."
        else:
            self.reorderPredicate(predicate_id, ndx - 1)
            msg = "Predicate+moved."
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_predicates'
                                     + '?manage_tabs_message=%s' % msg)

    @security.protected(ManagePortal)
    def doMovePredicateDown(self, predicate_id, REQUEST):
        """
        """
        predicate_ids = list(self.predicate_ids)
        ndx = predicate_ids.index(predicate_id)
        if ndx == len(predicate_ids) - 1:
            msg = "Predicate+already+last."
        else:
            self.reorderPredicate(predicate_id, ndx + 1)
            msg = "Predicate+moved."
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_predicates'
                                     + '?manage_tabs_message=%s' % msg)

    @security.protected(ManagePortal)
    def doRemovePredicate(self, predicate_id, REQUEST):
        """
        """
        self.removePredicate(predicate_id)
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                   + '/manage_predicates'
                                   + '?manage_tabs_message=Predicate+removed.')

    security.declareProtected(ManagePortal, 'manage_testRegistry')
    manage_testRegistry = DTMLFile('registryTest', _dtmldir)

    @security.protected(ManagePortal)
    def doTestRegistry(self, name, content_type, body, REQUEST):
        """
        """
        typeName = self.findTypeName(name, content_type, body)
        if typeName is None:
            typeName = '<unknown>'
        else:
            ttool = getUtility(ITypesTool)
            typeName = ttool.getTypeInfo(typeName).Title()
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_testRegistry'
                                     + '?testResults=Type:+%s'
                                     % urllib.parse.quote(typeName))

    #
    #   Predicate manipulation
    #
    security.declarePublic('getPredicate')
    def getPredicate(self, predicate_id):
        """
            Find the predicate whose id is 'id';  return the predicate
            object, if found, or else None.
        """
        return self.predicates.get(predicate_id, (None, None))[0]

    security.declarePublic('listPredicates')
    def listPredicates(self):
        """List '(id, (predicate, typeObjectName))' tuples for all predicates.
        """
        return tuple([ (id, self.predicates[id])
                       for id in self.predicate_ids ])

    security.declarePublic('getTypeObjectName')
    def getTypeObjectName(self, predicate_id):
        """
            Find the predicate whose id is 'id';  return the name of
            the type object, if found, or else None.
        """
        return self.predicates.get(predicate_id, (None, None))[1]

    @security.protected(ManagePortal)
    def addPredicate(self, predicate_id, predicate_type):
        """
            Add a predicate to this element of type 'typ' to the registry.
        """
        if predicate_id in self.predicate_ids:
            raise ValueError("Existing predicate: %s" % predicate_id)

        klass = None
        for key, value in _predicate_types:
            if key == predicate_type:
                klass = value

        if klass is None:
            raise ValueError("Unknown predicate type: %s" % predicate_type)

        self.predicates[predicate_id] = (klass(predicate_id), None)
        self.predicate_ids = self.predicate_ids + (predicate_id,)

    @security.protected(ManagePortal)
    def updatePredicate(self, predicate_id, predicate, typeObjectName):
        """
            Update a predicate in this element.
        """
        if not predicate_id in self.predicate_ids:
            raise ValueError("Unknown predicate: %s" % predicate_id)

        predObj = self.predicates[predicate_id][0]
        mapply(predObj.edit, (), predicate.__dict__)
        self.assignTypeName(predicate_id, typeObjectName)

    @security.protected(ManagePortal)
    def removePredicate(self, predicate_id):
        """
            Remove a predicate from the registry.
        """
        del self.predicates[predicate_id]
        idlist = list(self.predicate_ids)
        ndx = idlist.index(predicate_id)
        idlist = idlist[:ndx] + idlist[ndx + 1:]
        self.predicate_ids = tuple(idlist)

    @security.protected(ManagePortal)
    def reorderPredicate(self, predicate_id, newIndex):
        """
            Move a given predicate to a new location in the list.
        """
        idlist = list(self.predicate_ids)
        ndx = idlist.index(predicate_id)
        pred = idlist[ndx]
        idlist = idlist[:ndx] + idlist[ndx + 1:]
        idlist.insert(newIndex, pred)
        self.predicate_ids = tuple(idlist)

    @security.protected(ManagePortal)
    def assignTypeName(self, predicate_id, typeObjectName):
        """
            Bind the given predicate to a particular type object.
        """
        pred, _oldTypeObjName = self.predicates[predicate_id]
        self.predicates[predicate_id] = (pred, typeObjectName)

    #
    #   ContentTypeRegistry interface
    #
    def findTypeName(self, name, typ, body):
        """
            Perform a lookup over a collection of rules, returning the
            the name of the Type object corresponding to name/typ/body.
            Return None if no match found.
        """
        for predicate_id in self.predicate_ids:
            pred, typeObjectName = self.predicates[predicate_id]
            if pred(name, typ, body):
                return typeObjectName

        return None

InitializeClass(ContentTypeRegistry)
registerToolInterface('content_type_registry', IContentTypeRegistry)


def manage_addRegistry(self, REQUEST=None):
    """
        Add a CTR to self.
    """
    CTRID = ContentTypeRegistry.id
    reg = ContentTypeRegistry()
    self._setObject(CTRID, reg)
    reg = self._getOb(CTRID)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_main'
                                     + '?manage_tabs_message=Registry+added.')



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for ContentTypeRegistry module.
"""

import unittest
import Testing

from zope.interface.verify import verifyClass


class MajorMinorPredicateTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ContentTypeRegistry import MajorMinorPredicate

        return MajorMinorPredicate(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ContentTypeRegistry import MajorMinorPredicate
        from Products.CMFCore.interfaces import IContentTypeRegistryPredicate

        verifyClass(IContentTypeRegistryPredicate, MajorMinorPredicate)

    def test_empty(self):
        pred = self._makeOne('empty')
        self.assertEqual(pred.getMajorType(), 'None')
        self.assertEqual(pred.getMinorType(), 'None')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))

    def test_simple(self):
        pred = self._makeOne('plaintext')
        pred.edit('text', 'plain')
        self.assertEqual(pred.getMajorType(), 'text')
        self.assertEqual(pred.getMinorType(), 'plain')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo', 'text/html', 'asdfljksadf'))
        self.assertFalse(pred('', '', ''))
        self.assertFalse(pred('', 'asdf', ''))

    def test_wildcard(self):
        pred = self._makeOne('alltext')
        pred.edit('text', '')
        self.assertEqual(pred.getMajorType(), 'text')
        self.assertEqual(pred.getMinorType(), '')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo', 'text/html', 'asdfljksadf'))
        self.assertFalse(pred('foo', 'image/png', 'asdfljksadf'))

        pred.edit('', 'html')
        self.assertEqual(pred.getMajorType(), '')
        self.assertEqual(pred.getMinorType(), 'html')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo', 'text/html', 'asdfljksadf'))
        self.assertFalse(pred('foo', 'image/png', 'asdfljksadf'))


class ExtensionPredicateTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ContentTypeRegistry import ExtensionPredicate

        return ExtensionPredicate(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ContentTypeRegistry import ExtensionPredicate
        from Products.CMFCore.interfaces import IContentTypeRegistryPredicate

        verifyClass(IContentTypeRegistryPredicate, ExtensionPredicate)

    def test_empty(self):
        pred = self._makeOne('empty')
        self.assertEqual(pred.getExtensions(), 'None')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo.txt', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo.bar', 'text/html', 'asdfljksadf'))

    def test_simple(self):
        pred = self._makeOne('stardottext')
        pred.edit('txt')
        self.assertEqual(pred.getExtensions(), 'txt')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo.txt', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo.bar', 'text/html', 'asdfljksadf'))

    def test_multi(self):
        pred = self._makeOne('stardottext')
        pred.edit('txt text html htm')
        self.assertEqual(pred.getExtensions(), 'txt text html htm')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo.txt', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo.text', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo.html', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo.htm', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo.bar', 'text/html', 'asdfljksadf'))


class MimeTypeRegexPredicateTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ContentTypeRegistry \
                import MimeTypeRegexPredicate

        return MimeTypeRegexPredicate(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ContentTypeRegistry \
                import MimeTypeRegexPredicate
        from Products.CMFCore.interfaces import IContentTypeRegistryPredicate

        verifyClass(IContentTypeRegistryPredicate, MimeTypeRegexPredicate)

    def test_empty(self):
        pred = self._makeOne('empty')
        self.assertEqual(pred.getPatternStr(), 'None')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))

    def test_simple(self):
        pred = self._makeOne('plaintext')
        pred.edit('text/plain')
        self.assertEqual(pred.getPatternStr(), 'text/plain')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('foo', 'text/html', 'asdfljksadf'))

    def test_pattern(self):
        pred = self._makeOne('alltext')
        pred.edit('text/*')
        self.assertEqual(pred.getPatternStr(), 'text/*')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('foo', 'text/html', 'asdfljksadf'))
        self.assertFalse(pred('foo', 'image/png', 'asdfljksadf'))


class NameRegexPredicateTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ContentTypeRegistry import NameRegexPredicate

        return NameRegexPredicate(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ContentTypeRegistry import NameRegexPredicate
        from Products.CMFCore.interfaces import IContentTypeRegistryPredicate

        verifyClass(IContentTypeRegistryPredicate, NameRegexPredicate)

    def test_empty(self):
        pred = self._makeOne('empty')
        self.assertEqual(pred.getPatternStr(), 'None')
        self.assertFalse(pred('foo', 'text/plain', 'asdfljksadf'))

    def test_simple(self):
        pred = self._makeOne('onlyfoo')
        pred.edit('foo')
        self.assertEqual(pred.getPatternStr(), 'foo')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('fargo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('bar', 'text/plain', 'asdfljksadf'))

    def test_pattern(self):
        pred = self._makeOne('allfwords')
        pred.edit('f.*')
        self.assertEqual(pred.getPatternStr(), 'f.*')
        self.assertTrue(pred('foo', 'text/plain', 'asdfljksadf'))
        self.assertTrue(pred('fargo', 'text/plain', 'asdfljksadf'))
        self.assertFalse(pred('bar', 'text/plain', 'asdfljksadf'))


class ContentTypeRegistryTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ContentTypeRegistry import ContentTypeRegistry

        return ContentTypeRegistry(*args, **kw)

    def setUp(self):
        self.reg = self._makeOne()

    def test_interfaces(self):
        from Products.CMFCore.ContentTypeRegistry import ContentTypeRegistry
        from Products.CMFCore.interfaces import IContentTypeRegistry

        verifyClass(IContentTypeRegistry, ContentTypeRegistry)

    def test_empty(self):
        reg = self.reg
        self.assertTrue(reg.findTypeName('foo', 'text/plain', 'asdfljksadf')
                        is None)
        self.assertTrue(reg.findTypeName('fargo', 'text/plain', 'asdfljksadf')
                        is None)
        self.assertTrue(reg.findTypeName('bar', 'text/plain', 'asdfljksadf')
                        is None)
        self.assertFalse(reg.listPredicates())
        self.assertRaises(KeyError, reg.removePredicate, 'xyzzy')

    def test_reorder(self):
        reg = self.reg
        predIDs = ('foo', 'bar', 'baz', 'qux')
        for predID in predIDs:
            reg.addPredicate(predID, 'name_regex')
        ids = tuple([x[0] for x in reg.listPredicates()])
        self.assertEqual(ids, predIDs)
        reg.reorderPredicate('bar', 3)
        ids = tuple([x[0] for x in reg.listPredicates()])
        self.assertEqual(ids, ('foo', 'baz', 'qux', 'bar'))

    def test_lookup(self):
        reg = self.reg
        reg.addPredicate('image', 'major_minor')
        reg.getPredicate('image').edit('image', '')
        reg.addPredicate('onlyfoo', 'name_regex')
        reg.getPredicate('onlyfoo').edit('foo')
        reg.assignTypeName('onlyfoo', 'Foo')
        self.assertEqual(reg.findTypeName('foo', 'text/plain', 'asdfljksadf'),
                         'Foo')
        self.assertFalse(reg.findTypeName('fargo', 'text/plain',
                                          'asdfljksadf'))
        self.assertFalse(reg.findTypeName('bar', 'text/plain', 'asdfljksadf'))
        self.assertEqual(reg.findTypeName('foo', '', ''), 'Foo')
        self.assertEqual(reg.findTypeName('foo', None, None), 'Foo')


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(MajorMinorPredicateTests),
        unittest.makeSuite(ExtensionPredicateTests),
        unittest.makeSuite(MimeTypeRegexPredicateTests),
        unittest.makeSuite(NameRegexPredicateTests),
        unittest.makeSuite(ContentTypeRegistryTests),
        ))



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""TypeInformation browser views.
"""

from xml.dom.minidom import parseString

from zope.component import queryMultiAdapter
from zope.component import queryUtility

from Products.CMFCore.ActionInformation import Action
from Products.CMFCore.ActionInformation import ActionCategory
from Products.GenericSetup.browser.utils import AddWithPresettingsViewBase
from Products.GenericSetup.interfaces import INode
from Products.GenericSetup.interfaces import ISetupTool


class ActionAddView(AddWithPresettingsViewBase):

    """Add view for Action.
    """

    klass = Action

    description = u'An Action object represents a reference to an action.'

    def getProfileInfos(self):
        profiles = []
        stool = queryUtility(ISetupTool)
        if stool:
            for info in stool.listContextInfos():
                obj_ids = []
                context = stool._getImportContext(info['id'])
                body = context.readDataFile('actions.xml')
                if body is None:
                    continue
                root = parseString(body).documentElement
                for node in root.childNodes:
                    if node.nodeName != 'object':
                        continue
                    obj_ids += self._extractChildren(node)
                profiles.append({'id': info['id'],
                                 'title': info['title'],
                                 'obj_ids': tuple(sorted(obj_ids))})
        return tuple(profiles)

    def _extractChildren(self, node):
        action_paths = []
        category_id = node.getAttribute('name')
        for child in node.childNodes:
            if child.nodeName != 'object':
                continue
            if child.getAttribute('meta_type') == self.klass.meta_type:
                action_id = child.getAttribute('name')
                action_paths.append(action_id)
            else:
                action_paths += self._extractChildren(child)
        return [ ('%s/%s' % (category_id, path)) for path in action_paths ]

    def _initSettings(self, obj, profile_id, obj_path):
        stool = queryUtility(ISetupTool)
        if stool is None:
            return

        context = stool._getImportContext(profile_id)
        body = context.readDataFile('actions.xml')
        if body is None:
            return

        settings_node = None
        root = parseString(body).documentElement
        for node in root.childNodes:
            if node.nodeName != 'object':
                continue
            for obj_id in obj_path:
                for child in node.childNodes:
                    if child.nodeName != 'object':
                        continue
                    if child.getAttribute('name') != obj_id:
                        continue
                    if child.getAttribute('meta_type') == self.klass.meta_type:
                        settings_node = child
                    else:
                        node = child
                    break

        importer = queryMultiAdapter((obj, context), INode)
        if importer is None:
            return

        importer.node = settings_node
        return


class ActionCategoryAddView(AddWithPresettingsViewBase):

    """Add view for ActionCategory.
    """

    klass = ActionCategory

    description = \
            u'An Action Category object represents a group of Action objects.'

    def getProfileInfos(self):
        return []

    def _initSettings(self, obj, profile_id, obj_path):
        pass



import os
from setuptools import setup
from setuptools import find_packages

NAME = 'CMFCore'

here = os.path.abspath(os.path.dirname(__file__))
package = os.path.join(here, 'Products', NAME)


def _package_doc(name):
    f = open(os.path.join(here, name))
    return f.read()


_boundary = '\n' + ('-' * 60) + '\n\n'
README = _boundary.join([
    _package_doc('README.txt'),
    _package_doc('CHANGES.txt'),
])

setup(name='Products.%s' % NAME,
      version='2.4.0b4',
      description='Zope Content Management Framework core components',
      long_description=README,
      classifiers=[
          "Development Status :: 4 - Beta",
          "Framework :: Plone",
          "Framework :: Zope :: 4",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Zope Public License",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: Implementation :: CPython",
          "Topic :: Software Development :: Libraries :: Application Frameworks",  # noqa
      ],
      keywords='web application server zope cmf',
      author="Zope Foundation and Contributors",
      author_email="zope-cmf@zope.org",
      url="https://github.com/zopefoundation/Products.CMFCore",
      license="ZPL 2.1",
      packages=find_packages(),
      include_package_data=True,
      namespace_packages=['Products'],
      zip_safe=False,
      setup_requires=[
          'eggtestinfo',
      ],
      install_requires=[
          'setuptools',
          'Zope >= 4.0b4',
          'docutils',
          'five.localsitemanager',
          'Products.BTreeFolder2',
          'Products.GenericSetup >= 2.0b1',
          'Products.MailHost >= 4.0',
          'Products.PythonScripts',
          'Products.StandardCacheManagers',
          'Products.ZCTextIndex',
          'six',
          ],
      tests_require=[
          'zope.testing >= 3.7.0',
          'Products.StandardCacheManagers',
          ],
      extras_require={
          'test': ['Products.StandardCacheManagers'],
          'zsql': ['Products.ZSQLMethods >= 3.0.0b1'],
          },
      test_loader='zope.testing.testrunner.eggsupport:SkipLayers',
      test_suite='Products.%s' % NAME,
      entry_points="""
      [zope2.initialize]
      Products.%s = Products.%s:initialize
      [distutils.commands]
      ftest = zope.testing.testrunner.eggsupport:ftest
      """ % (NAME, NAME),
      )



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for CachingPolicyManager module.
"""

import unittest
import Testing

import base64
import os
from os.path import join as path_join

from AccessControl.SecurityManagement import newSecurityManager
from Acquisition import Implicit
from App.Common import rfc1123_date
from DateTime.DateTime import DateTime
from OFS.Cache import Cacheable
from zope.component import getSiteManager
from zope.interface.verify import verifyClass

from Products.CMFCore.FSDTMLMethod import FSDTMLMethod
from Products.CMFCore.FSPageTemplate import FSPageTemplate
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.interfaces import ITypesTool
from Products.CMFCore.testing import FunctionalZCMLLayer
from Products.CMFCore.testing import TraversingZCMLLayer
from Products.CMFCore.tests.base.dummy import DummyContent
from Products.CMFCore.tests.base.dummy import DummySite
from Products.CMFCore.tests.base.dummy import DummyTool
from Products.CMFCore.tests.base.testcase import FSDVTest
from Products.CMFCore.tests.base.testcase import SecurityTest
from Products.CMFCore.tests.base.testcase import TransactionalTest

ACCLARK = DateTime('2001/01/01')
portal_owner = b'portal_owner'


class DummyContent2:

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, modified):
        self.modified = modified

    def Type(self):
        return 'Dummy'

    def modified(self):
        return self.modified


class CacheableDummyContent(Implicit, Cacheable):

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, id):
        self.id = id
        self.modified = DateTime()

    def getId(self):
        """ """
        return self.id

    def modified(self):
        return self.modified

    def __call__(self):
        """ """
        if self.ZCacheable_isCachingEnabled():
            result = self.ZCacheable_get(default=None)
            if result is not None:
                # We will always get None from RAMCacheManager and HTTP
                # Accelerated Cache Manager but we will get
                # something implementing the IStreamIterator interface
                # from a "FileCacheManager"
                return result

        self.ZCacheable_set(None)


class DummyView(CacheableDummyContent):

    meta_type = 'DTML Method'


class CachingPolicyTests(unittest.TestCase):

    layer = TraversingZCMLLayer

    def _makePolicy(self, policy_id, **kw):
        from Products.CMFCore.CachingPolicyManager import CachingPolicy

        return CachingPolicy(policy_id, **kw)

    def _makeContext(self, **kw):
        from Products.CMFCore.CachingPolicyManager import createCPContext

        return createCPContext(DummyContent2(self._epoch),
                               'foo_view', kw, self._epoch)

    def setUp(self):
        self._epoch = DateTime(0)
        getSiteManager().registerUtility(DummyTool(), IMembershipTool)

    def test_interfaces(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicy
        from Products.CMFCore.interfaces import ICachingPolicy

        verifyClass(ICachingPolicy, CachingPolicy)

    def test_empty(self):
        policy = self._makePolicy('empty')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], 'Last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

    def test_noPassPredicate(self):
        policy = self._makePolicy('noPassPredicate', predicate='nothing')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

    def test_typePredicate(self):
        policy = self._makePolicy('typePredicate',
                                  predicate='python:object.Type() == "Dummy"')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], 'Last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

    def test_typePredicateMiss(self):
        policy = self._makePolicy('typePredicate',
                                 predicate='python:object.Type() == "Foolish"')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

    def test_viewPredicate(self):
        policy = self._makePolicy('viewPredicate',
                                  predicate='python:view == "foo_view"')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], 'Last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

    def test_viewPredicateMiss(self):
        policy = self._makePolicy('viewPredicateMiss',
                                  predicate='python:view == "bar_view"')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

    def test_kwPredicate(self):
        policy = self._makePolicy('kwPredicate',
                                  predicate='python:"foo" in keywords.keys()')
        context = self._makeContext(foo=1)
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], 'Last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

    def test_kwPredicateMiss(self):
        policy = self._makePolicy('kwPredicateMiss',
                                  predicate='python:"foo" in keywords.keys()')
        context = self._makeContext(bar=1)
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

    def test_mtimeFunc(self):
        policy = self._makePolicy('mtimeFunc', mtime_func='string:2001/01/01')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0], 'Last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(ACCLARK.timeTime()))

    def test_mtimeFuncNone(self):
        policy = self._makePolicy('mtimeFuncNone', mtime_func='nothing')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)

    def test_maxAge(self):
        policy = self._makePolicy('aged', max_age_secs=86400)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'expires')
        self.assertEqual(headers[1][1],
                         rfc1123_date((self._epoch + 1).timeTime()))
        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'max-age=86400')

    def test_sMaxAge(self):
        policy = self._makePolicy('s_aged', s_max_age_secs=86400)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 's-maxage=86400')
        self.assertEqual(policy.getSMaxAgeSecs(), 86400)

    def test_noCache(self):
        policy = self._makePolicy('noCache', no_cache=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'pragma')
        self.assertEqual(headers[1][1], 'no-cache')
        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'no-cache')

    def test_noStore(self):
        policy = self._makePolicy('noStore', no_store=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'no-store')

    def test_mustRevalidate(self):
        policy = self._makePolicy('mustRevalidate', must_revalidate=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'must-revalidate')

    def test_proxyRevalidate(self):
        policy = self._makePolicy('proxyRevalidate', proxy_revalidate=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'proxy-revalidate')
        self.assertEqual(policy.getProxyRevalidate(), 1)

    def test_public(self):
        policy = self._makePolicy('public', public=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'public')
        self.assertEqual(policy.getPublic(), 1)

    def test_private(self):
        policy = self._makePolicy('private', private=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'private')
        self.assertEqual(policy.getPrivate(), 1)

    def test_noTransform(self):
        policy = self._makePolicy('noTransform', no_transform=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'no-transform')
        self.assertEqual(policy.getNoTransform(), 1)

    def test_lastModified(self):
        policy = self._makePolicy('lastModified', last_modified=0)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 0)
        self.assertEqual(policy.getLastModified(), 0)

    def test_preCheck(self):
        policy = self._makePolicy('preCheck', pre_check=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'pre-check=1')
        self.assertEqual(policy.getPreCheck(), 1)
        self.assertEqual(policy.getPostCheck(), None)

    def test_postCheck(self):
        policy = self._makePolicy('postCheck', post_check=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'cache-control')
        self.assertEqual(headers[1][1], 'post-check=1')
        self.assertEqual(policy.getPostCheck(), 1)
        self.assertEqual(policy.getPreCheck(), None)

    def test_ETag(self):
        # With an empty etag_func, no ETag should be produced
        policy = self._makePolicy('ETag', etag_func='')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

        policy = self._makePolicy('ETag', etag_func='string:foo')
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 2)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'etag')
        self.assertEqual(headers[1][1], 'foo')

    def test_combined(self):
        policy = self._makePolicy('noStore', no_cache=1, no_store=1)
        context = self._makeContext()
        headers = policy.getHeaders(context)

        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'pragma')
        self.assertEqual(headers[1][1], 'no-cache')
        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'no-cache, no-store')


class CachingPolicyManagerTests(unittest.TestCase):

    layer = TraversingZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicyManager

        return CachingPolicyManager

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def setUp(self):
        self._epoch = DateTime()
        getSiteManager().registerUtility(DummyTool(), IMembershipTool)

    def assertEqualDelta(self, lhs, rhs, delta):
        self.assertTrue(abs(lhs - rhs) <= delta)

    def test_interfaces(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicyManager
        from Products.CMFCore.interfaces import ICachingPolicyManager

        verifyClass(ICachingPolicyManager, CachingPolicyManager)

    def test_empty(self):
        mgr = self._makeOne()

        self.assertEqual(len(mgr.listPolicies()), 0)
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={},
                                            time=self._epoch)
        self.assertEqual(len(headers), 0)

        self.assertRaises(KeyError, mgr._updatePolicy,
                           'xyzzy', None, None, None, None, None, None, '',
                           '', None, None, None, None, None)
        self.assertRaises(KeyError, mgr._removePolicy, 'xyzzy')
        self.assertRaises(KeyError, mgr._reorderPolicy, 'xyzzy', -1)

    def test_addAndUpdatePolicy(self):
        mgr = self._makeOne()
        mgr.addPolicy('first', 'python:1', 'mtime', 1, 0, 1, 0, 'vary',
                      'etag', None, 2, 1, 0, 1, 0, 1, 0, 2, 3)
        p = mgr._policies['first']
        self.assertEqual(p.getPolicyId(), 'first')
        self.assertEqual(p.getPredicate(), 'python:1')
        self.assertEqual(p.getMTimeFunc(), 'mtime')
        self.assertEqual(p.getMaxAgeSecs(), 1)
        self.assertEqual(p.getNoCache(), 0)
        self.assertEqual(p.getNoStore(), 1)
        self.assertEqual(p.getMustRevalidate(), 0)
        self.assertEqual(p.getVary(), 'vary')
        self.assertEqual(p.getETagFunc(), 'etag')
        self.assertEqual(p.getSMaxAgeSecs(), 2)
        self.assertEqual(p.getProxyRevalidate(), 1)
        self.assertEqual(p.getPublic(), 0)
        self.assertEqual(p.getPrivate(), 1)
        self.assertEqual(p.getNoTransform(), 0)
        self.assertEqual(p.getEnable304s(), 1)
        self.assertEqual(p.getLastModified(), 0)
        self.assertEqual(p.getPreCheck(), 2)
        self.assertEqual(p.getPostCheck(), 3)

        mgr.updatePolicy('first', 'python:0', 'mtime2', 2, 1, 0, 1, 'vary2',
                         'etag2', None, 1, 0, 1, 0, 1, 0, 1, 3, 2)
        p = mgr._policies['first']
        self.assertEqual(p.getPolicyId(), 'first')
        self.assertEqual(p.getPredicate(), 'python:0')
        self.assertEqual(p.getMTimeFunc(), 'mtime2')
        self.assertEqual(p.getMaxAgeSecs(), 2)
        self.assertEqual(p.getNoCache(), 1)
        self.assertEqual(p.getNoStore(), 0)
        self.assertEqual(p.getMustRevalidate(), 1)
        self.assertEqual(p.getVary(), 'vary2')
        self.assertEqual(p.getETagFunc(), 'etag2')
        self.assertEqual(p.getSMaxAgeSecs(), 1)
        self.assertEqual(p.getProxyRevalidate(), 0)
        self.assertEqual(p.getPublic(), 1)
        self.assertEqual(p.getPrivate(), 0)
        self.assertEqual(p.getNoTransform(), 1)
        self.assertEqual(p.getEnable304s(), 0)
        self.assertEqual(p.getLastModified(), 1)
        self.assertEqual(p.getPreCheck(), 3)
        self.assertEqual(p.getPostCheck(), 2)

    def test_reorder(self):
        mgr = self._makeOne()

        policy_ids = ('foo', 'bar', 'baz', 'qux')

        for policy_id in policy_ids:
            mgr._addPolicy(policy_id,
                           'python:"%s" in keywords.keys()' % policy_id,
                           None, 0, 0, 0, 0, '', '')

        ids = tuple([x[0] for x in mgr.listPolicies()])
        self.assertEqual(ids, policy_ids)

        mgr._reorderPolicy('bar', 3)

        ids = tuple([x[0] for x in mgr.listPolicies()])
        self.assertEqual(ids, ('foo', 'baz', 'qux', 'bar'))

    def _makeOneWithPolicies(self):
        mgr = self._makeOne()

        policy_tuples = (('foo', None),
                         ('bar', 0),
                         ('baz', 3600),
                         ('qux', 86400))

        for policy_id, max_age_secs in policy_tuples:
            mgr._addPolicy(policy_id,
                           'python:"%s" in keywords.keys()' % policy_id,
                           None, max_age_secs, 0, 0, 0, '', '')

        return mgr

    def test_lookupNoMatch(self):
        mgr = self._makeOneWithPolicies()
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={},
                                            time=self._epoch)
        self.assertEqual(len(headers), 0)

    def test_lookupMatchFoo(self):
        mgr = self._makeOneWithPolicies()
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={'foo': 1},
                                            time=self._epoch)
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))

    def test_lookupMatchBar(self):
        mgr = self._makeOneWithPolicies()
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={'bar': 1},
                                            time=self._epoch)
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'expires')
        self.assertEqual(headers[1][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'max-age=0')

    def test_lookupMatchBaz(self):
        mgr = self._makeOneWithPolicies()
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={'baz': 1},
                                            time=self._epoch)
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'expires')

        exp_time = DateTime(headers[1][1])
        target = self._epoch + (1.0 / 24.0)
        self.assertEqualDelta(exp_time, target, 0.01)

        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'max-age=3600')

    def test_lookupMatchQux(self):
        mgr = self._makeOneWithPolicies()
        headers = mgr.getHTTPCachingHeaders(content=DummyContent2(self._epoch),
                                            view_method='foo_view',
                                            keywords={'qux': 1},
                                            time=self._epoch)
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers[0][0].lower(), 'last-modified')
        self.assertEqual(headers[0][1], rfc1123_date(self._epoch.timeTime()))
        self.assertEqual(headers[1][0].lower(), 'expires')

        exp_time = DateTime(headers[1][1])
        target = self._epoch + 1.0
        self.assertEqualDelta(exp_time, target, 0.01)

        self.assertEqual(headers[2][0].lower(), 'cache-control')
        self.assertEqual(headers[2][1], 'max-age=86400')


class CachingPolicyManager304Tests(SecurityTest, FSDVTest):

    layer = TraversingZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicyManager

        return CachingPolicyManager

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def setUp(self):
        from Products.CMFCore.interfaces import ICachingPolicyManager

        SecurityTest.setUp(self)
        FSDVTest.setUp(self)

        now = DateTime()

        # Create a fake portal and the tools we need
        self.portal = DummySite(id='portal').__of__(self.app)
        cpm = self._makeOne()
        sm = getSiteManager()
        sm.registerUtility(cpm, ICachingPolicyManager)
        sm.registerUtility(DummyTool(), IMembershipTool)
        sm.registerUtility(DummyTool(), ITypesTool)

        # This is a FSPageTemplate that will be used as the View for
        # our content objects. It doesn't matter what it returns.
        path = os.path.join(self.skin_path_name, 'testPT2.pt')
        self.portal._setObject('dummy_view',
                               FSPageTemplate('dummy_view', path))

        uf = self.app.acl_users
        password = b'secret'
        uf.userFolderAddUser(portal_owner, password, ['Manager'], [])
        user = uf.getUserById(portal_owner)
        if not hasattr(user, 'aq_base'):
            user = user.__of__(uf)
        newSecurityManager(None, user)
        owner_auth = b'%s:%s' % (portal_owner, password)
        self.auth_header = "Basic %s" % base64.encodestring(owner_auth)

        self.portal._setObject('doc1', DummyContent('doc1'))
        self.portal._setObject('doc2', DummyContent('doc2'))
        self.portal._setObject('doc3', DummyContent('doc3'))
        self.portal.doc1.modified_date = now
        self.portal.doc2.modified_date = now
        self.portal.doc3.modified_date = now

        # This policy only applies to doc1. It will not emit any ETag header
        # but it enables If-modified-since handling.
        cpm.addPolicy(policy_id='policy_no_etag',
                      predicate='python:object.getId()=="doc1"',
                      mtime_func='',
                      max_age_secs=0,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='',
                      etag_func='',
                      enable_304s=1)

        # This policy only applies to doc2. It will emit an ETag with
        # the constant value "abc" and also enable if-modified-since handling.
        cpm.addPolicy(policy_id='policy_etag',
                      predicate='python:object.getId()=="doc2"',
                      mtime_func='',
                      max_age_secs=0,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='',
                      etag_func='string:abc',
                      enable_304s=1)

        # This policy only applies to doc3. Etags with constant values of
        # "abc" are emitted, but if-modified-since handling is turned off.
        cpm.addPolicy(policy_id='policy_disabled',
                      predicate='python:object.getId()=="doc3"',
                      mtime_func='',
                      max_age_secs=0,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='',
                      etag_func='string:abc',
                      enable_304s=0)

    def tearDown(self):
        FSDVTest.tearDown(self)
        SecurityTest.tearDown(self)

    def _cleanup(self):
        # Clean up request and response
        req = self.portal.REQUEST

        for header in ('IF_MODIFIED_SINCE', 'HTTP_AUTHORIZATION',
                       'IF_NONE_MATCH'):
            if req.environ.get(header, None) is not None:
                del req.environ[header]

        req.RESPONSE.setStatus(200)

    def testUnconditionalGET(self):
        # In this case the Request does not specify any if-modified-since
        # value to take into account, thereby completely circumventing any
        # if-modified-since handling. This must not produce a response status
        # of 304, regardless of any other headers.
        self.portal.doc1()
        response = self.portal.REQUEST.RESPONSE
        self.assertEqual(response.getStatus(), 200)

    def testConditionalGETNoETag(self):
        yesterday = DateTime() - 1
        doc1 = self.portal.doc1
        request = doc1.REQUEST
        response = request.RESPONSE

        # If doc1 has beeen modified since yesterday (which it has), we want
        # the full rendering.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(yesterday)
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc1()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # If doc1 has been modified since its creation (which it hasn't), we
        # want the full rendering. This must return a 304 response.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc1.modified_date)
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc1()
        self.assertEqual(response.getStatus(), 304)
        self._cleanup()

        # ETag handling is not enabled in the policy for doc1, so asking for
        # one will not produce any matches. We get the full rendering.
        request.environ['IF_NONE_MATCH'] = '"123"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc1()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # We are asking for an ETag as well as modifications after doc2 has
        # been created. Both won't match and wwe get the full rendering.
        request.environ['IF_NONE_MATCH'] = '"123"'
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc1.modified_date)
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc1()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

    def testConditionalGETETag(self):
        yesterday = DateTime() - 1
        doc2 = self.portal.doc2
        request = doc2.REQUEST
        response = request.RESPONSE

        # Has doc2 been modified since yesterday? Yes it has, so we get the
        # full rendering.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(yesterday)
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # If doc2 has not been modified since its creation (which it hasn't),
        # we would get a 304 here. However, the policy for doc2 also expects
        # to get an ETag in the request, which we are not setting here. So
        # the policy fails and we get a full rendering.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc2.modified_date)
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # Now we are setting an ETag in our request, but an ETag that does not
        # match the policy's expected value. The policy fails and we get the
        # full rendering.
        request.environ['IF_NONE_MATCH'] = '"123"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # Here we provide the correct and matching ETag value, and we don't
        # specify any if-modified-since condition. This is enough for our
        # policy to trigger 304.
        request.environ['IF_NONE_MATCH'] = '"abc"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 304)
        self._cleanup()

        # We specify an ETag and a modification time condition that dooes not
        # match, so we get the full rendering
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc2.modified_date)
        request.environ['IF_NONE_MATCH'] = '"123"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # We hand in a matching modified time condition which is supposed to
        # trigger full rendering. This will lead the ETag condition to be
        # overrridden.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(yesterday)
        request.environ['IF_NONE_MATCH'] = '"abc"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # Now we pass an ETag that matches the policy and a modified time
        # condition that is not fulfilled. It is safe to serve a 304.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc2.modified_date)
        request.environ['IF_NONE_MATCH'] = '"abc"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc2()
        self.assertEqual(response.getStatus(), 304)
        self._cleanup()

    def testConditionalGETDisabled(self):
        doc3 = self.portal.doc3
        request = doc3.REQUEST
        response = request.RESPONSE

        # Our policy disables any 304-handling, so even though the ETag matches
        # the policy, we will get the full rendering.
        request.environ['IF_NONE_MATCH'] = '"abc"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc3()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()

        # Now both the ETag and the modified condition would trigger a 304
        # response *if* 304-handling was enabled. It is not in our policy, so
        # we get the full rendering again.
        request.environ['IF_MODIFIED_SINCE'] = rfc1123_date(doc3.modified_date)
        request.environ['IF_NONE_MATCH'] = '"abc"'
        request.environ['HTTP_AUTHORIZATION'] = self.auth_header
        doc3()
        self.assertEqual(response.getStatus(), 200)
        self._cleanup()


class FSObjMaker(FSDVTest):

    def _makeFSPageTemplate(self, id, filename):
        path = path_join(self.skin_path_name, filename)
        return FSPageTemplate(id, path)

    def _makeFSDTMLMethod(self, id, filename):
        path = path_join(self.skin_path_name, filename)
        return FSDTMLMethod(id, path)


class NestedTemplateTests(TransactionalTest, FSObjMaker):

    layer = TraversingZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicyManager

        return CachingPolicyManager

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def setUp(self):
        from Products.CMFCore.interfaces import ICachingPolicyManager

        FSObjMaker.setUp(self)
        TransactionalTest.setUp(self)

        # Create a fake portal and the tools we need
        self.portal = DummySite(id='portal').__of__(self.app)
        self.cpm = self._makeOne()
        sm = getSiteManager()
        sm.registerUtility(self.cpm, ICachingPolicyManager)
        sm.registerUtility(DummyTool(), IMembershipTool)
        sm.registerUtility(DummyTool(), ITypesTool)

    def tearDown(self):
        TransactionalTest.tearDown(self)
        FSObjMaker.tearDown(self)

    def test_subtemplate_cpm_1(self):
        # test that subtemplates dont call the cpm
        # set up site
        portal = self.portal
        now = DateTime()
        cpm = self.cpm
        cpm.addPolicy(policy_id='policy_op2',
                      predicate='python:view=="output_page_2"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      s_max_age_secs=100
                      )

        content = DummyContent(id='content', view_id='output_page_1')
        content.modified_date = now
        portal._setObject('content', content)

        output_page_1 = self._makeFSPageTemplate('output_page_1',
                                                 'output_page_1.zpt')
        output_page_2 = self._makeFSPageTemplate('output_page_2',
                                                 'output_page_2.zpt')
        portal._setObject('output_page_1', output_page_1)
        portal._setObject('output_page_2', output_page_2)

        portal.content()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_subtemplate_cpm_2(self):
        # test that calling content from a template doesnt call the cpm
        # just calling an FSDTMLMethod directly from another template does
        # not activate the bug because RESPONSE is not passed in
        portal = self.portal
        now = DateTime()
        cpm = self.cpm
        cpm.addPolicy(policy_id='policy_op4',
                      predicate='python:view=="output_page_4"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      s_max_age_secs=100
                      )

        content = DummyContent(id='content', view_id='output_page_3')
        content.modified_date = now
        portal._setObject('content', content)
        content2 = DummyContent(id='content2', view_id='output_page_4')
        content2.modified_date = now
        portal._setObject('content2', content2)

        output_page_3 = self._makeFSDTMLMethod('output_page_3',
                                               'output_page_3.dtml')
        output_page_4 = self._makeFSDTMLMethod('output_page_4',
                                               'output_page_4.dtml')
        portal._setObject('output_page_4', output_page_4)
        portal._setObject('output_page_3', output_page_3)

        # call the content
        portal.content()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_subtemplate_cpm_3(self):
        # test a bigger mix of zpt templates
        # set up site
        portal = self.portal
        now = DateTime()
        cpm = self.cpm
        cpm.addPolicy(policy_id='policy_nv1',
                      predicate='python:view=="nested_view_1"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      s_max_age_secs=100
                      )

        doc1 = DummyContent(id='doc1', view_id='nested_view')
        doc1.modified_date = now
        portal._setObject('doc1', doc1)
        doc2 = DummyContent(id='doc2', view_id='nested_view_1')
        doc2.modified_date = now
        portal._setObject('doc2', doc2)
        doc3 = DummyContent(id='doc3', view_id='nested_view_2')
        doc3.modified_date = now
        portal._setObject('doc3', doc3)

        nested_view = self._makeFSPageTemplate('nested_view',
                                               'nested_view.zpt')
        nested_view_1 = self._makeFSPageTemplate('nested_view_1',
                                                 'nested_view_1.zpt')
        nested_view_2 = self._makeFSPageTemplate('nested_view_2',
                                                 'nested_view_2.zpt')
        portal._setObject('nested_view', nested_view)
        portal._setObject('nested_view_1', nested_view_1)
        portal._setObject('nested_view_2', nested_view_2)

        portal.doc1()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_mixed_subtemplate_cpm(self):
        # test a mix of zpt and dtml templates
        # set up site
        now = DateTime()
        portal = self.portal
        cpm = self.cpm
        cpm.addPolicy(policy_id='policy_nv1',
                      predicate='python:view=="nested_view_1"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      s_max_age_secs=100
                      )

        doc1 = DummyContent(id='doc1', view_id='nested_view',
                            modified_date=now)
        portal._setObject('doc1', doc1)
        doc2 = DummyContent(id='doc2', view_id='nested_view_1',
                            modified_date=now)
        portal._setObject('doc2', doc2)
        doc3 = DummyContent(id='doc3', view_id='nested_view_2',
                            modified_date=now)
        portal._setObject('doc3', doc3)

        nested_view = self._makeFSPageTemplate('nested_view',
                                               'nested_view.zpt')
        nested_view_1 = self._makeFSPageTemplate('nested_view_1',
                                                 'nested_view_1.zpt')
        nested_view_2 = self._makeFSDTMLMethod('nested_view_2',
                                               'nested_view_2.dtml')
        portal._setObject('nested_view', nested_view)
        portal._setObject('nested_view_1', nested_view_1)
        portal._setObject('nested_view_2', nested_view_2)

        portal.doc1()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_fireForSubtemplates(self):
        # This is a FSPageTemplate that will be used as the View for
        # our content objects. It doesn't matter what it returns.
        dv = self._makeFSPageTemplate('dummy_view', 'testPT_CPM1.zpt')
        self.portal._setObject('dummy_view', dv)

        # These are the subtemplates we use
        sv1 = self._makeFSPageTemplate('subview_1', 'testPT_CPM2.zpt')
        sv2 = self._makeFSDTMLMethod('subview_2', 'testDTML_CPM3.dtml')
        self.portal._setObject('subview_1', sv1)
        self.portal._setObject('subview_2', sv2)

        for i in (1, 2, 3):
            id = 'doc%i' % i
            title = 'Document %i' % i
            description = 'This is document %i' % i
            modified_date = DateTime()
            doc = DummyContent(id)
            doc.title = title
            doc.description = description
            doc.modified_date = modified_date
            self.portal._setObject(id, doc)

        cpm = self.cpm

        # This policy only applies to doc2.
        cpm.addPolicy(policy_id='policy_doc2',
                      predicate='python:object.getId()=="doc2"',
                      mtime_func='',
                      max_age_secs=200,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc2',
                      etag_func='',
                      pre_check=1
                      )

        # This policy only applies to doc3.
        cpm.addPolicy(policy_id='policy_doc3',
                      predicate='python:object.getId()=="doc3"',
                      mtime_func='',
                      max_age_secs=300,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc3',
                      etag_func='',
                      post_check=1
                      )

        # http://www.zope.org/Collectors/CMF/456
        # In cases where one view (ZPT or DTML) is rendered from another
        # view, we want to ensure only the view requested by the visitor
        # will get caching rules applied.
        self.portal.doc1.dummy_view()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_fireForSubtemplates2(self):
        # This is a FSPageTemplate that will be used as the View for
        # our content objects. It doesn't matter what it returns.
        dv = self._makeFSPageTemplate('dummy_view', 'testPT_CPM1.zpt')
        self.portal._setObject('dummy_view', dv)

        # These are the subtemplates we use
        sv1 = self._makeFSPageTemplate('subview_1', 'testPT_CPM2.zpt')
        sv2 = self._makeFSDTMLMethod('subview_2', 'testDTML_CPM3.dtml')
        self.portal._setObject('subview_1', sv1)
        self.portal._setObject('subview_2', sv2)

        for i in (1, 2, 3):
            id = 'doc%i' % i
            title = 'Document %i' % i
            description = 'This is document %i' % i
            modified_date = DateTime()
            doc = DummyContent(id)
            doc.title = title
            doc.description = description
            doc.modified_date = modified_date
            self.portal._setObject(id, doc)

        cpm = self.cpm

        # This policy only applies to doc1.
        cpm.addPolicy(policy_id='policy_doc1',
                      predicate='python:object.getId()=="doc1"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      s_max_age_secs=100
                      )

        # This policy only applies to doc2.
        cpm.addPolicy(policy_id='policy_doc2',
                      predicate='python:object.getId()=="doc2"',
                      mtime_func='',
                      max_age_secs=200,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc2',
                      etag_func='',
                      pre_check=1
                      )

        # This policy only applies to doc3.
        cpm.addPolicy(policy_id='policy_doc3',
                      predicate='python:object.getId()=="doc3"',
                      mtime_func='',
                      max_age_secs=300,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc3',
                      etag_func='',
                      post_check=1
                      )

        # http://www.zope.org/Collectors/CMF/456
        # In cases where one view (ZPT or DTML) is rendered from another
        # view, we want to ensure only the view requested by the visitor
        # will get caching rules applied.
        self.portal.doc1.dummy_view()

        # We want to make sure the correct policy (policy_doc1) has fired
        # Just to be sure, change headers so they are definitely all
        # lower-cased
        headers = {}
        header_info = self.RESPONSE.headers.items()
        [headers.__setitem__(x[0].lower(), x[1]) for x in header_info]

        self.assertTrue(headers.get('x-cache-headers-set-by'))
        self.assertEqual(headers.get('vary'), 'doc1')
        self.assertEqual(headers.get('cache-control'),
                         'max-age=100, s-maxage=100')


class OFSCacheTests(TransactionalTest):

    layer = FunctionalZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.CachingPolicyManager import CachingPolicyManager

        return CachingPolicyManager

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def setUp(self):
        from Products.CMFCore.interfaces import ICachingPolicyManager

        TransactionalTest.setUp(self)

        # Create a fake portal and the tools we need
        self.portal = DummySite(id='portal').__of__(self.app)
        self.portal._setObject('doc1', CacheableDummyContent('doc1'))
        self.portal._setObject('doc2', CacheableDummyContent('doc2'))
        cpm = self._makeOne()
        self.portal._setObject('caching_policy_manager', cpm)
        sm = getSiteManager()
        sm.registerUtility(cpm, ICachingPolicyManager)
        sm.registerUtility(DummyTool(), IMembershipTool)

        # This policy only applies to doc1. It will not emit any ETag header
        # but it enables If-modified-since handling.
        cpm.addPolicy(policy_id='policy_1',
                      predicate='python:object.getId()=="doc1"',
                      mtime_func='',
                      max_age_secs=100,
                      no_cache=0,
                      no_store=0,
                      must_revalidate=0,
                      vary='doc1',
                      etag_func='',
                      enable_304s=0)

    def tearDown(self):
        from Products.CMFCore.interfaces import ICachingPolicyManager

        sm = getSiteManager()
        sm.unregisterUtility(provided=ICachingPolicyManager)
        sm.unregisterUtility(provided=IMembershipTool)
        TransactionalTest.tearDown(self)

    def test_empty(self):
        from Products.CMFCore.CachingPolicyManager import CPMCache

        cpm = self.portal.caching_policy_manager
        doc1 = self.portal.doc1
        self.assertTrue(cpm._isCacheManager)
        self.assertTrue(isinstance(cpm.ZCacheManager_getCache(), CPMCache))
        self.assertEqual(doc1.ZCacheable_getManagerIds(),
                         ({'id': cpm.getId(), 'title': ''},))

    def test_no_association(self):
        # Render an item that would match the CPM policy, but don't
        # associate it with the CPM.
        self.portal.doc1()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_unsuitable_association(self):
        # Render an item that is associated with the CPM, but that does not
        # match any policy.
        cpm = self.portal.caching_policy_manager
        doc2 = self.portal.doc2
        doc2.ZCacheable_setManagerId(cpm.getId())

        doc2()

        # no headers should be added by the CPM if all is well
        headers = [x.lower() for x in self.RESPONSE.headers.keys()]
        self.assertFalse('x-cache-headers-set-by' in headers)
        self.assertFalse('vary' in headers)

    def test_suitable_association(self):
        # Render a content item that will trigger the CPM
        cpm = self.portal.caching_policy_manager
        doc1 = self.portal.doc1
        doc1.ZCacheable_setManagerId(cpm.getId())

        doc1()

        # Policy "policy_1" should have triggered
        # Just to be sure, change headers so they are definitely all
        # lower-cased
        headers = {}
        header_info = self.RESPONSE.headers.items()
        [headers.__setitem__(x[0].lower(), x[1]) for x in header_info]

        self.assertTrue(headers.get('x-cache-headers-set-by'))
        self.assertEqual(headers.get('vary'), 'doc1')
        self.assertEqual(headers.get('cache-control'), 'max-age=100')

    def test_with_view(self):
        # Render a view for a content item that will trigger the CPM
        cpm = self.portal.caching_policy_manager
        self.portal._setObject('a_view', DummyView(id='a_view'))
        self.portal.a_view.ZCacheable_setManagerId(cpm.getId())
        doc1 = self.portal.doc1

        doc1.a_view()

        # Policy "policy_1" should have triggered
        # Just to be sure, change headers so they are definitely all
        # lower-cased
        headers = {}
        header_info = self.RESPONSE.headers.items()
        [headers.__setitem__(x[0].lower(), x[1]) for x in header_info]

        self.assertTrue(headers.get('x-cache-headers-set-by'))
        self.assertEqual(headers.get('vary'), 'doc1')
        self.assertEqual(headers.get('cache-control'), 'max-age=100')


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(CachingPolicyTests),
        unittest.makeSuite(CachingPolicyManagerTests),
        unittest.makeSuite(CachingPolicyManager304Tests),
        unittest.makeSuite(NestedTemplateTests),
        unittest.makeSuite(OFSCacheTests),
        ))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for WorkflowTool module.
"""

import unittest
import Testing

from OFS.SimpleItem import SimpleItem
from zope.component import adapter
from zope.component import getSiteManager
from zope.component import provideHandler
from zope.interface import implementer
from zope.interface.verify import verifyClass
from zope.testing.cleanup import cleanUp

from Products.CMFCore.interfaces import IActionRaisedExceptionEvent
from Products.CMFCore.interfaces import IActionSucceededEvent
from Products.CMFCore.interfaces import IActionWillBeInvokedEvent
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import ITypesTool
from Products.CMFCore.interfaces import IWorkflowAware
from Products.CMFCore.interfaces import IWorkflowDefinition
from Products.CMFCore.testing import TraversingEventZCMLLayer


class Dummy(SimpleItem):

    def __init__(self, id):
        self._id = id

    def getId(self):
        return self._id


@implementer(IWorkflowDefinition)
class DummyWorkflow(Dummy):

    meta_type = 'DummyWorkflow'
    _isAWorkflow = 1
    _known_actions = ()
    _known_info = ()

    def __init__(self, id):
        Dummy.__init__(self, id)
        self._did_action = {}
        self._gave_info = {}
        self._notified = {}

    def setKnownActions(self, known_actions):
        self._known_actions = known_actions

    def setKnownInfo(self, known_info):
        self._known_info = known_info

    def didAction(self, action):
        return self._did_action.setdefault(action, [])

    def gaveInfo(self, name):
        return self._gave_info.setdefault(name, [])

    def notified(self, name):
        return self._notified.setdefault(name, [])

    #
    #   WorkflowDefinition interface
    #
    def getCatalogVariablesFor(self, ob):
        return {'dummy': '%s: %s' % (self.getId(), ob.getId())}

    def updateRoleMappingsFor(self, ob):
        pass

    def listObjectActions(self, info):
        return () #XXX

    def listGlobalActions(self, info):
        return () #XXX

    def isActionSupported(self, ob, action):
        return action in self._known_actions

    def doActionFor(self, ob, action, *args, **kw):
        self.didAction(action).append(ob)

    def isInfoSupported(self, ob, name):
        return name in self._known_info

    def getInfoFor(self, ob, name, default, *args, **kw):
        self.gaveInfo(name).append(ob)
        return name in self._known_info and 1 or 0

    def notifyCreated(self, ob):
        self.notified('created').append((ob,))

    def notifyBefore(self, ob, action):
        self.notified('before').append((ob, action))

    def notifySuccess(self, ob, action, result):
        self.notified('success').append((ob, action, result))

    def notifyException(self, ob, action, exc):
        self.notified('exception').append((ob, action, exc))

@adapter(IActionWillBeInvokedEvent)
def notifyBeforeHandler(evt):
    evt.workflow.notified('before-evt').append((evt.object, evt.action))

@adapter(IActionSucceededEvent)
def notifySuccessHandler(evt):
    evt.workflow.notified('success-evt').append((evt.object, evt.action,
                                                 evt.result))

@adapter(IActionRaisedExceptionEvent)
def notifyExceptionHandler(evt):
    evt.workflow.notified('exception-evt').append((evt.object, evt.action,
                                                   evt.exc))


@implementer(IContentish, IWorkflowAware)
class DummyContent(Dummy):

    meta_type = 'Dummy'

    def getPortalTypeName(self):
        return 'Dummy Content'


class DummyNotReallyContent(Dummy):

    meta_type = 'Dummy Content'


class DummyTypeInfo(Dummy):

    pass


class DummyTypesTool(SimpleItem):

    def listTypeInfo(self):
        return [DummyTypeInfo('Dummy Content')]

    def getTypeInfo(self, ob):
        if getattr(ob, 'meta_type', None) is 'Dummy':
            return DummyTypeInfo('Dummy Content')
        return None


class WorkflowToolTests(unittest.TestCase):

    layer = TraversingEventZCMLLayer

    def _makeOne(self, workflow_ids=()):
        from Products.CMFCore.WorkflowTool import WorkflowTool

        tool = WorkflowTool()

        for workflow_id in workflow_ids:
            tool._setObject(workflow_id, DummyWorkflow(workflow_id))

        return tool

    def _makeRoot(self):
        from OFS.Folder import Folder

        root = Folder('root')
        getSiteManager().registerUtility(DummyTypesTool(), ITypesTool)
        return root

    def _makeWithTypes(self):
        root = self._makeRoot()
        return self._makeOne(workflow_ids=('a', 'b')).__of__(root)

    def _makeWithTypesAndChain(self):
        tool = self._makeWithTypes()
        tool.setChainForPortalTypes(('Dummy Content',), ('a', 'b'))
        return tool

    def tearDown(self):
        cleanUp()

    def test_interfaces(self):
        from Products.CMFCore.interfaces import IActionProvider
        from Products.CMFCore.interfaces import IConfigurableWorkflowTool
        from Products.CMFCore.interfaces import IWorkflowTool
        from Products.CMFCore.WorkflowTool import WorkflowTool

        verifyClass(IActionProvider, WorkflowTool)
        verifyClass(IConfigurableWorkflowTool, WorkflowTool)
        verifyClass(IWorkflowTool, WorkflowTool)

    def test_empty(self):
        from Products.CMFCore.WorkflowTool import WorkflowException

        tool = self._makeOne()

        self.assertFalse(tool.getWorkflowIds())
        self.assertEqual(tool.getWorkflowById('default_workflow'), None)
        self.assertEqual(tool.getWorkflowById('a'), None)

        self.assertRaises(WorkflowException, tool.getInfoFor, None, 'hmm')
        self.assertRaises(WorkflowException, tool.doActionFor, None, 'hmm')

    def test_new_with_wf(self):
        from Products.CMFCore.WorkflowTool import WorkflowException

        tool = self._makeWithTypes()

        wfids = tool.getWorkflowIds()
        self.assertEqual(len(wfids), 2)
        self.assertTrue('a' in wfids)
        self.assertTrue('b' in wfids)
        self.assertEqual(tool.getWorkflowById('default'), None)
        wf = tool.getWorkflowById('a')
        self.assertEqual(wf.getId(), 'a')
        wf = tool.getWorkflowById('b')
        self.assertEqual(wf.getId(), 'b')

        self.assertRaises(WorkflowException, tool.getInfoFor, None, 'hmm')
        self.assertRaises(WorkflowException, tool.doActionFor, None, 'hmm')

    def test_nonContent(self):
        tool = self._makeWithTypesAndChain()
        self.assertEqual(len(tool.getDefaultChainFor(None)), 0)
        self.assertEqual(len(tool.getDefaultChain()), 1)
        self.assertEqual(len(tool.listChainOverrides()), 1)
        self.assertEqual(len(tool.getChainFor(None)), 0)
        self.assertEqual(len(tool.getCatalogVariablesFor(None)), 0)

    def test_notReallyContent(self):
        tool = self._makeWithTypesAndChain()
        dummy = DummyNotReallyContent('doh')
        self.assertEqual(len(tool.getDefaultChainFor(dummy)), 0)
        self.assertEqual(len(tool.getDefaultChain()), 1)
        self.assertEqual(len(tool.listChainOverrides()), 1)
        self.assertEqual(len(tool.getChainFor(dummy)), 0)
        self.assertEqual(len(tool.getCatalogVariablesFor(dummy)), 0)

    def test_content_default_chain(self):
        tool = self._makeWithTypes()
        dummy = DummyContent('dummy')
        self.assertEqual(len(tool.getDefaultChainFor(dummy)), 1)
        self.assertEqual(len(tool.getDefaultChain()), 1)
        self.assertEqual(len(tool.listChainOverrides()), 0)
        self.assertEqual(len(tool.getChainFor(dummy)), 1)
        self.assertEqual(len(tool.getCatalogVariablesFor(dummy)), 0)
        self.assertEqual(tool.getDefaultChain(), tool.getChainFor(dummy))

    def test_content_own_chain(self):
        tool = self._makeWithTypesAndChain()

        dummy = DummyContent('dummy')

        self.assertEqual(len(tool.getDefaultChainFor(dummy)), 1)
        self.assertEqual(len(tool.getDefaultChain()), 1)
        self.assertEqual(len(tool.listChainOverrides()), 1)
        chain = tool.getChainFor(dummy)
        self.assertEqual(len(chain), 2)
        self.assertTrue('a' in chain)
        self.assertTrue('b' in chain)

        vars = tool.getCatalogVariablesFor(dummy)
        self.assertEqual(len(vars), 1)
        self.assertTrue('dummy' in vars.keys())
        self.assertTrue('a: dummy' in vars.values())

    def test_setChainForPortalTypes(self):
        tool = self._makeWithTypes()
        tool.setDefaultChain('b, a')
        dummy = DummyContent('dummy')

        tool.setChainForPortalTypes(('Dummy Content',), ('a', 'b'))
        self.assertEqual(tool.getChainFor(dummy), ('a', 'b'))
        tool.setChainForPortalTypes(('Dummy Content',), 'a, b')
        self.assertEqual(tool.getChainFor(dummy), ('a', 'b'))

        tool.setChainForPortalTypes(('Dummy Content',), ())
        self.assertEqual(tool.getChainFor(dummy), ())
        tool.setChainForPortalTypes(('Dummy Content',), '')
        self.assertEqual(tool.getChainFor(dummy), ())

        tool.setChainForPortalTypes(('Dummy Content',), None)
        self.assertEqual(tool.getChainFor(dummy), ('b', 'a'))

        # Using the '(Default)' keyword
        # https://bugs.launchpad.net/zope-cmf/+bug/161702
        tool.setChainForPortalTypes(('Dummy Content',), '(Default)')
        self.assertEqual(tool.getDefaultChain(), tool.getChainFor(dummy))
        tool.setDefaultChain('a, b')
        self.assertEqual(tool.getDefaultChain(), tool.getChainFor(dummy))

    def test_getCatalogVariablesFor(self):
        tool = self._makeWithTypesAndChain()
        dummy = DummyContent('dummy')

        vars = tool.getCatalogVariablesFor(dummy)
        self.assertEqual(len(vars), 1)
        self.assertTrue('dummy' in vars.keys())
        self.assertTrue('a: dummy' in vars.values())

    def test_getInfoFor(self):
        tool = self._makeWithTypesAndChain()
        tool.b.setKnownInfo(('info',))
        dummy = DummyContent('dummy')

        info = tool.getInfoFor(dummy, 'info')

        self.assertEqual(info, 1)
        self.assertFalse(tool.a.gaveInfo('info'))
        self.assertTrue(tool.b.gaveInfo('info'))

    def test_doActionFor(self):
        tool = self._makeWithTypesAndChain()
        tool.a.setKnownActions(('action',))
        dummy = DummyContent('dummy')

        tool.doActionFor(dummy, 'action')

        self.assertTrue(tool.a.didAction('action'))
        self.assertFalse(tool.b.didAction('action'))

    def test_notifyCreated(self):
        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        tool.notifyCreated(ob)

        for wf in tool.a, tool.b:
            notified = wf.notified('created')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob,))

    def test_notifyCreated_skips_workflows_with_status(self):
        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        dummy_status0 = object()
        ob.workflow_history = {'a': (dummy_status0,)}
        tool.notifyCreated(ob)

        notified = tool.a.notified('created')
        self.assertEqual(len(notified), 0)

        notified = tool.b.notified('created')
        self.assertEqual(len(notified), 1)
        self.assertEqual(notified[0], (ob,))

    def test_notifyBefore(self):
        provideHandler(notifyBeforeHandler)

        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        tool.notifyBefore(ob, 'action')

        for wf in tool.a, tool.b:
            notified = wf.notified('before')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action'))

            notified = wf.notified('before-evt')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action'))

    def test_notifySuccess(self):
        provideHandler(notifySuccessHandler)

        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        tool.notifySuccess(ob, 'action')

        for wf in tool.a, tool.b:
            notified = wf.notified('success')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action', None))

            notified = wf.notified('success-evt')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action', None))

    def test_notifyException(self):
        provideHandler(notifyExceptionHandler)

        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        tool.notifyException(ob, 'action', 'exception')

        for wf in tool.a, tool.b:
            notified = wf.notified('exception')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action', 'exception'))

            notified = wf.notified('exception-evt')
            self.assertEqual(len(notified), 1)
            self.assertEqual(notified[0], (ob, 'action', 'exception'))

    def test_getStatusOf(self):
        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        dummy_status0 = object()
        dummy_status1 = object()
        ob.workflow_history = {'a': (dummy_status0, dummy_status1)}

        self.assertEqual(tool.getStatusOf('a', ob), dummy_status1)

    def test_setStatusOf(self):
        tool = self._makeWithTypesAndChain()

        ob = DummyContent('dummy')
        dummy_status0 = object()
        tool.setStatusOf('a', ob, dummy_status0)

        self.assertTrue('a' in ob.workflow_history)
        self.assertEqual(len(ob.workflow_history['a']), 1)
        self.assertEqual(ob.workflow_history['a'][0], dummy_status0)

        dummy_status1 = object()
        tool.setStatusOf('a', ob, dummy_status1)

        self.assertTrue('a' in ob.workflow_history)
        self.assertEqual(len(ob.workflow_history['a']), 2)
        self.assertEqual(ob.workflow_history['a'][0], dummy_status0)
        self.assertEqual(ob.workflow_history['a'][1], dummy_status1)

    def xxx_test_updateRoleMappings(self):
        """
            Build a tree of objects, invoke tool.updateRoleMappings,
            and then check to see that the workflows each got called;
            check the resulting count, as well.
        """


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(WorkflowToolTests),
        ))



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Content type registry xml adapters and setup handlers.
"""

from zope.component import adapts
from zope.component import getSiteManager

from Products.GenericSetup.interfaces import ISetupEnviron
from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects
from Products.GenericSetup.utils import XMLAdapterBase

from Products.CMFCore.interfaces import IContentTypeRegistry


class ContentTypeRegistryXMLAdapter(XMLAdapterBase):

    """XML im- and exporter for ContentTypeRegistry.
    """

    adapts(IContentTypeRegistry, ISetupEnviron)

    _LOGGER_ID = 'contenttypes'

    name = 'contenttyperegistry'

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._getObjectNode('object')
        node.appendChild(self._extractPredicates())

        self._logger.info('Content type registry exported.')
        return node

    def _importNode(self, node):
        """Import the object from the DOM node.
        """
        if self.environ.shouldPurge():
            self._purgePredicates()

        self._initPredicates(node)

        self._logger.info('Content type registry imported.')

    def _extractPredicates(self):
        fragment = self._doc.createDocumentFragment()
        for predicate_id, info in self.context.listPredicates():
            child = self._doc.createElement('predicate')
            child.setAttribute('name', predicate_id)
            child.setAttribute('predicate_type', info[0].PREDICATE_TYPE)
            child.setAttribute('content_type_name', info[1])
            for argument in self._crackArgs(info[0]):
                sub = self._doc.createElement('argument')
                sub.setAttribute('value', argument)
                child.appendChild(sub)
            fragment.appendChild(child)
        return fragment

    def _purgePredicates(self):
        self.context.__init__()

    def _initPredicates(self, node):
        for child in node.childNodes:
            if child.nodeName != 'predicate':
                continue
            parent = self.context

            predicate_id = str(child.getAttribute('name'))
            if predicate_id not in parent.predicate_ids:
                predicate_type = str(child.getAttribute('predicate_type'))
                parent.addPredicate(predicate_id, predicate_type)

            if child.hasAttribute('content_type_name'):
                content_type_name = str(child.getAttribute('content_type_name'))
                parent.assignTypeName(predicate_id, content_type_name)

            if child.hasAttribute('insert-before'):
                insert_before = child.getAttribute('insert-before')
                self._movePredicate(predicate_id, insert_before, 0)
            elif child.hasAttribute('insert-after'):
                insert_after = child.getAttribute('insert-after')
                self._movePredicate(predicate_id, insert_after, 1)

            arguments = []
            for sub in child.childNodes:
                if sub.nodeName != 'argument':
                    continue
                arguments.append(str(sub.getAttribute('value')))
            if arguments:
                parent.getPredicate(predicate_id).edit(*arguments)

    _KNOWN_PREDICATE_TYPES = {
        'major_minor': lambda x: (','.join(x.major or ()),
                                  ','.join(x.minor or ())),
        'extension': lambda x: (','.join(x.extensions or ()),),
        'mimetype_regex': lambda x: (x.pattern and x.pattern.pattern or '',),
        'name_regex': lambda x: (x.pattern and x.pattern.pattern or '',),
    }

    def _crackArgs(self, predicate):
        cracker = self._KNOWN_PREDICATE_TYPES.get(predicate.PREDICATE_TYPE)
        if cracker is not None:
            return cracker(predicate)
        return ()  # XXX:  raise?

    def _movePredicate(self, id, position_id, delta=0):
        predicate_ids = list(self.context.predicate_ids)
        if position_id == '*':
            position_id = predicate_ids[-delta]
        if position_id == id:
            return
        try:
            predicate_ids.remove(id)
            position = predicate_ids.index(position_id)
            predicate_ids.insert(position + delta, id)
        except ValueError:
            return
        self.context.predicate_ids = tuple(predicate_ids)

def importContentTypeRegistry(context):
    """Import content type registry settings from an XML file.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IContentTypeRegistry)
    if tool is None:
        logger = context.getLogger('contenttypes')
        logger.debug('Nothing to import.')
        return

    importObjects(tool, '', context)

def exportContentTypeRegistry(context):
    """Export content type registry settings as an XML file.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IContentTypeRegistry)
    if tool is None:
        logger = context.getLogger('contenttypes')
        logger.debug('Nothing to export.')
        return

    exportObjects(tool, '', context)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for zcml module.
"""

import doctest
import unittest


def test_registerDirectory():
    """
    Use the cmf:registerDirectory directive::

      >>> import Products.CMFCore
      >>> from Zope2.App import zcml
      >>> configure_zcml = '''
      ... <configure xmlns:cmf="http://namespaces.zope.org/cmf">
      ...   <cmf:registerDirectory
      ...       name="fake_skin"
      ...       directory="tests/fake_skins/fake_skin"
      ...       recursive="True"
      ...       ignore="foo bar"
      ...       />
      ... </configure>'''
      >>> zcml.load_config('meta.zcml', Products.CMFCore)
      >>> zcml.load_string(configure_zcml)

    Make sure the directory is registered correctly::

      >>> from Products.CMFCore.DirectoryView import _dirreg
      >>> reg_keys = ('Products.CMFCore:tests/fake_skins/fake_skin',
      ...             'Products.CMFCore:tests/fake_skins/fake_skin/test_directory')
      >>> reg_keys[0] in _dirreg._directories
      True
      >>> reg_keys[1] in _dirreg._directories
      True
      >>> info = _dirreg._directories[reg_keys[0]]
      >>> info._reg_key == reg_keys[0]
      True
      >>> info.ignore
      ('.', '..', 'foo', 'bar')

    Clean up and make sure the cleanup works::

      >>> from zope.testing.cleanup import cleanUp
      >>> cleanUp()
      >>> reg_keys[0] in _dirreg._directories
      False
      >>> reg_keys[1] in _dirreg._directories
      False
    """


def test_suite():
    return unittest.TestSuite((
        doctest.DocTestSuite(),
        ))



DOCTYPE = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">'''

HTML_TEMPLATE = '''\
<html><head>
 <title>%(title)s</title>
</head>
<body bgcolor="#efe843">%(body)s</body>
</html>
'''

SIMPLE_HTML = '''\
<html>
 <head>
  <title>Title in tag</title>
  <meta name="description" content="Describe me">
  <meta name="contributors" content="foo@bar.com; baz@bam.net;
    Benotz, Larry J (larry@benotz.stuff)">
  <meta name="title" content="Title in meta">
  <meta name="subject" content="content management">
 </head>
 <body bgcolor="#ffffff">
  <h1>Not a lot here</h1>
 </body>
</html>
'''

BASIC_HTML = '''\
<html>
 <head>
  <title>Title in tag</title>
  <meta name="description" content="Describe me">
  <meta name="contributors" content="foo@bar.com; baz@bam.net;
    Benotz, Larry J (larry@benotz.stuff)">
  <meta name="title" content="Title in meta">
  <meta name="subject" content="content management">
  <meta name="keywords" content="unit tests, framework; ,zope ">
 </head>
 <body bgcolor="#ffffff">
  <h1>Not a lot here</h1>
 </body>
</html>
'''

SIMPLE_XHTML = """\
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html
    PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>Title in tag</title>
  <meta name="description" content="Describe me" />
  <meta name="contributors" content="foo@bar.com; baz@bam.net;
    Benotz, Larry J (larry@benotz.stuff)" />
  <meta name="title" content="Title in meta" />
  <meta name="subject" content="content management" />
 </head>
 <body bgcolor="#ffffff">
  <h1>Not a lot here</h1>
 </body>
</html>
"""

# A document with an html-qualifying *portion*.
FAUX_HTML_LEADING_TEXT = '''\
The following would look like HTML but for this leading text:

<html>
 <head>
  <title>Title in tag</title>
  <meta name="description" content="Describe me">
  <meta name="contributors" content="foo@bar.com; baz@bam.net;
    Benotz, Larry J (larry@benotz.stuff)">
  <meta name="title" content="Title in meta">
  <meta name="subject" content="content management">
  <meta name="keywords" content="unit tests, framework; ,zope ">
 </head>
 <body bgcolor="#ffffff">
  <h1>Not a lot here</h1>
 </body>
</html>
'''

ENTITY_IN_TITLE = '''\
<html>
 <head>
  <title>&Auuml;rger</title>
 </head>
 <bOdY>
  <h2>Not a lot here either</h2>
 </bodY>
</html>
'''

SIMPLE_STRUCTUREDTEXT = '''\
Title: My Document
Description: A document by me
Contributors: foo@bar.com; baz@bam.net; no@yes.maybe
Subject: content management, zope

This is the header

  Body body body body body
  body body body.

   o A list item

   o And another thing...
'''

BASIC_STRUCTUREDTEXT = '''\
Title: My Document
Description: A document by me
Contributors: foo@bar.com; baz@bam.net; no@yes.maybe
Subject: content management, zope
Keywords: unit tests; , framework

This is the header

  Body body body body body
  body body body.

   o A list item

   o And another thing...
'''

STX_WITH_HTML = """\
Sometimes people do interesting things

  Sometimes people do interesting things like have examples
  of HTML inside their structured text document.  We should
  be detecting that this is indeed a structured text document
  and **NOT** an HTML document::

    <html>
    <head><title>Hello World</title></head>
    <body><p>Hello world, I am Bruce.</p></body>
    </html>

  All in favor say pi!
"""


STX_NO_HEADERS = """\
Title Phrase

    This is a "plain" STX file, with no headers.  Saving with
    it shouldn't overwrite any metadata.
"""

STX_NO_HEADERS_BUT_COLON = """\
Plain STX:  No magic!

    This is a "plain" STX file, with no headers.  Saving with
    it shouldn't overwrite any metadata.
"""

BASIC_RFC822 = """\
Title: Zope Community
Description: Link to the Zope Community website.
Subject: open source; Zope; community

http://www.zope.org
"""

RFC822_W_CONTINUATION = """\
Title: Zope Community
Description: Link to the Zope Community website,
  including hundreds of contributed Zope products.
Subject: open source; Zope; community

http://www.zope.org
"""

BASIC_ReST = '''\
Title: My Document
Description: A document by me
Contributors: foo@bar.com; baz@bam.net; no@yes.maybe
Subject: content management, zope
Keywords: unit tests; , framework

==================
This is the header
==================

Body body body body body body body body.

* A list item

* And another thing...
'''

ReST_WITH_HTML = """\
Sometimes people do interesting things
======================================

Sometimes people do interesting things like have examples
of HTML inside their ReStructured text document.  We should
be detecting that this is indeed a restructured text document
and **NOT** an HTML document::

  <html>
  <head><title>Hello World</title></head>
  <body><p>Hello world, I am Bruce.</p></body>
  </html>

All in favor say pi!
"""


ReST_NO_HEADERS = """\
============
Title Phrase
============
This is a "plain" ReST file, with no headers.  Saving with
it shouldn't overwrite any metadata.
"""

ReST_NO_HEADERS_BUT_COLON = """\
======================
Plain ReST:  No magic!
======================

This is a "plain" ReST file, with no headers.  Saving with
it shouldn't overwrite any metadata.
"""



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Basic membership tool.
"""

import logging
import six
from warnings import warn

from AccessControl.class_init import InitializeClass
from AccessControl.requestmethod import postonly
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from AccessControl.User import nobody
from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.Dialogs import MessageDialog
from App.special_dtml import DTMLFile
from OFS.Folder import Folder
from Persistence import PersistentMapping
from ZODB.POSException import ConflictError
from zope.component import getUtility
from zope.component import queryUtility
from zope.component.interfaces import IFactory
from zope.globalrequest import getRequest
from zope.interface import implementedBy
from zope.interface import implementer
from ZPublisher.BaseRequest import RequestContainer

from Products.CMFCore.exceptions import AccessControl_Unauthorized
from Products.CMFCore.exceptions import BadRequest
from Products.CMFCore.interfaces import ICookieCrumbler
from Products.CMFCore.interfaces import IMemberDataTool
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.interfaces import IRegistrationTool
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.interfaces import ITypesTool
from Products.CMFCore.permissions import AccessContentsInformation
from Products.CMFCore.permissions import ChangeLocalRoles
from Products.CMFCore.permissions import ListPortalMembers
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.permissions import ManageUsers
from Products.CMFCore.permissions import SetOwnPassword
from Products.CMFCore.permissions import View
from Products.CMFCore.PortalFolder import PortalFolder
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.utils import _dtmldir
from Products.CMFCore.utils import Message as _
from Products.CMFCore.utils import registerToolInterface
from Products.CMFCore.utils import UniqueObject

logger = logging.getLogger('CMFCore.MembershipTool')


@implementer(IMembershipTool)
class MembershipTool(UniqueObject, Folder):

    """ This tool accesses member data through an acl_users object.

    It can be replaced with something that accesses member data in a
    different way.
    """

    id = 'portal_membership'
    meta_type = 'CMF Membership Tool'
    memberareaCreationFlag = 1
    _HOME_FOLDER_FACTORY_NAME = 'cmf.folder.home.bbb1'

    security = ClassSecurityInfo()

    manage_options = (
        ({'label': 'Configuration', 'action': 'manage_mapRoles'},
         {'label': 'Overview', 'action': 'manage_overview'}) +
        Folder.manage_options)

    #
    #   ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = DTMLFile('explainMembershipTool', _dtmldir)

    #
    #   'portal_membership' interface methods
    #
    security.declareProtected(ManagePortal, 'manage_mapRoles')
    manage_mapRoles = DTMLFile('membershipRolemapping', _dtmldir)

    @security.protected(SetOwnPassword)
    @postonly
    def setPassword(self, password, domains=None, REQUEST=None):
        '''Allows the authenticated member to set his/her own password.
        '''
        if not self.isAnonymousUser():
            member = self.getAuthenticatedMember()
            rtool = queryUtility(IRegistrationTool)
            if rtool is not None:
                failMessage = rtool.testPasswordValidity(password)
                if failMessage is not None:
                    raise BadRequest(failMessage)
            member.setSecurityProfile(password=password, domains=domains)
        else:
            raise BadRequest('Not logged in.')

    @security.public
    def getAuthenticatedMember(self):
        '''
        Returns the currently authenticated member object
        or the Anonymous User.  Never returns None.
        '''
        u = getSecurityManager().getUser()
        if u is None:
            u = nobody
        return self.wrapUser(u)

    @security.private
    def wrapUser(self, u, wrap_anon=0):
        """ Set up the correct acquisition wrappers for a user object.

        Provides an opportunity for a portal_memberdata tool to retrieve and
        store member data independently of the user object.
        """
        b = getattr(u, 'aq_base', None)
        if b is None:
            # u isn't wrapped at all.  Wrap it in self.acl_users.
            b = u
            u = u.__of__(self.acl_users)
        if (b is nobody and not wrap_anon) or hasattr(b, 'getMemberId'):
            # This user is either not recognized by acl_users or it is
            # already registered with something that implements the
            # member data tool at least partially.
            return u

        # Apply any role mapping if we have it
        if hasattr(self, 'role_map'):
            for portal_role in self.role_map.keys():
                if (self.role_map.get(portal_role) in u.roles and
                        portal_role not in u.roles):
                    u.roles.append(portal_role)

        mdtool = queryUtility(IMemberDataTool)
        if mdtool is not None:
            try:
                u = mdtool.wrapUser(u)
            except ConflictError:
                raise
            except:
                logger.exception("Error during wrapUser")
        return u

    @security.protected(ManagePortal)
    def getPortalRoles(self):
        """
        Return all local roles defined by the portal itself,
        which means roles that are useful and understood
        by the portal object
        """
        parent = self.aq_inner.aq_parent
        roles = list(parent.userdefined_roles())

        # This is *not* a local role in the portal but used by it
        roles.append('Manager')
        roles.append('Owner')

        return roles

    @security.protected(ManagePortal)
    @postonly
    def setRoleMapping(self, portal_role, userfolder_role, REQUEST=None):
        """
        set the mapping of roles between roles understood by
        the portal and roles coming from outside user sources
        """
        if not hasattr(self, 'role_map'):
            self.role_map = PersistentMapping()

        if len(userfolder_role) < 1:
            del self.role_map[portal_role]
        else:
            self.role_map[portal_role] = userfolder_role

        return MessageDialog(
               title='Mapping updated',
               message='The Role mappings have been updated',
               action='manage_mapRoles')

    @security.protected(ManagePortal)
    def getMappedRole(self, portal_role):
        """
        returns a role name if the portal role is mapped to
        something else or an empty string if it is not
        """
        if hasattr(self, 'role_map'):
            return self.role_map.get(portal_role, '')
        else:
            return ''

    @security.public
    def getMembersFolder(self):
        """ Get the members folder object.
        """
        parent = aq_parent(aq_inner(self))
        members_folder = getattr(parent, 'Members', None)
        if members_folder is None:
            return None
        request_container = RequestContainer(REQUEST=getRequest())
        return members_folder.__of__(request_container)

    @security.protected(ManagePortal)
    def getMemberareaCreationFlag(self):
        """
        Returns the flag indicating whether the membership tool
        will create a member area if an authenticated user from
        an underlying user folder logs in first without going
        through the join process
        """
        return self.memberareaCreationFlag

    @security.protected(ManagePortal)
    def setMemberareaCreationFlag(self):
        """
        sets the flag indicating whether the membership tool
        will create a member area if an authenticated user from
        an underlying user folder logs in first without going
        through the join process
        """
        if not hasattr(self, 'memberareaCreationFlag'):
            self.memberareaCreationFlag = 0

        if self.memberareaCreationFlag == 0:
            self.memberareaCreationFlag = 1
        else:
            self.memberareaCreationFlag = 0

        return MessageDialog(
               title='Member area creation flag changed',
               message='Member area creation flag has been updated',
               action='manage_mapRoles')

    @security.public
    def createMemberArea(self, member_id=''):
        """ Create a member area for 'member_id' or authenticated user.
        """
        if not self.getMemberareaCreationFlag():
            return None
        members = self.getMembersFolder()
        if members is None:
            return None
        if self.isAnonymousUser():
            return None
        if member_id:
            if not self.isMemberAccessAllowed(member_id):
                return None
            member = self.getMemberById(member_id)
            if member is None:
                return None
        else:
            member = self.getAuthenticatedMember()
            member_id = member.getId()
        if hasattr(aq_base(members), member_id):
            return None

        factory_name = self._HOME_FOLDER_FACTORY_NAME
        portal_type_name = 'Folder'
        ttool = queryUtility(ITypesTool)
        if ttool is not None:
            portal_type = ttool.getTypeInfo('Home Folder')
            if portal_type is not None:
                factory_name = portal_type.factory
                portal_type_name = portal_type.getId()

        factory = getUtility(IFactory, factory_name)
        obj = factory(id=member_id)
        obj._setPortalTypeName(portal_type_name)
        members._setObject(member_id, obj)
        f = members._getOb(member_id)
        f.changeOwnership(member)
        return f

    security.declarePublic('createMemberarea')
    createMemberarea = createMemberArea

    @security.protected(ManageUsers)
    @postonly
    def deleteMemberArea(self, member_id, REQUEST=None):
        """ Delete member area of member specified by member_id.
        """
        members = self.getMembersFolder()
        if not members:
            return 0
        if hasattr(aq_base(members), member_id):
            members.manage_delObjects(member_id)
            return 1
        else:
            return 0

    @security.public
    def isAnonymousUser(self):
        '''
        Returns 1 if the user is not logged in.
        '''
        u = getSecurityManager().getUser()
        if u is None or u.getUserName() == 'Anonymous User':
            return 1
        return 0

    @security.public
    def checkPermission(self, permissionName, object, subobjectName=None):
        '''
        Checks whether the current user has the given permission on
        the given object or subobject.
        '''
        if subobjectName is not None:
            object = getattr(object, subobjectName)
        return _checkPermission(permissionName, object)

    @security.protected(ManageUsers)
    def isMemberAccessAllowed(self, member_id):
        """Check if the authenticated user is this member or an user manager.
        """
        sm = getSecurityManager()
        user = sm.getUser()
        if user is None:
            return False
        if member_id == user.getId():
            return True
        return sm.checkPermission(ManageUsers, self)

    @security.public
    def credentialsChanged(self, password, REQUEST=None):
        '''
        Notifies the authentication mechanism that this user has changed
        passwords.  This can be used to update the authentication cookie.
        Note that this call should *not* cause any change at all to user
        databases.
        '''
        if not self.isAnonymousUser():
            user = getSecurityManager().getUser()
            name = user.getUserName()
            # this really does need to be the user name, and not the user id,
            # because we're dealing with authentication credentials
            cctool = queryUtility(ICookieCrumbler)
            if cctool is not None:
                cctool.credentialsChanged(user, name, password, REQUEST)

    @security.protected(ManageUsers)
    def getMemberById(self, id):
        '''
        Returns the given member.
        '''
        user = self._huntUser(id, self)
        if user is not None:
            user = self.wrapUser(user)
        return user

    def _huntUserFolder(self, member_id, context):
        """Find userfolder containing user in the hierarchy
           starting from context
        """
        uf = context.acl_users
        while uf is not None:
            user = uf.getUserById(member_id)
            if user is not None:
                return uf
            container = aq_parent(aq_inner(uf))
            parent = aq_parent(aq_inner(container))
            uf = getattr(parent, 'acl_users', None)
        return None

    def _huntUser(self, member_id, context):
        """Find user in the hierarchy of userfolders
           starting from context
        """
        uf = self._huntUserFolder(member_id, context)
        if uf is not None:
            return uf.getUserById(member_id).__of__(uf)

    def __getPUS(self):
        """ Retrieve the nearest user folder
        """
        warn('__getPUS is deprecated and will be removed in CMF 2.4, '
             'please acquire "acl_users" instead.', DeprecationWarning,
             stacklevel=2)
        return self.acl_users

    @security.protected(ManageUsers)
    def listMemberIds(self):
        '''Lists the ids of all members.  This may eventually be
        replaced with a set of methods for querying pieces of the
        list rather than the entire list at once.
        '''
        user_folder = self.acl_users
        return [ x.getId() for x in user_folder.getUsers() ]

    @security.protected(ManageUsers)
    def listMembers(self):
        '''Gets the list of all members.
        '''
        return list(map(self.wrapUser, self.acl_users.getUsers()))

    @security.protected(ListPortalMembers)
    def searchMembers(self, search_param, search_term):
        """ Search the membership """
        mdtool = queryUtility(IMemberDataTool)
        if mdtool is not None:
            return mdtool.searchMemberData(search_param, search_term)
        return None

    @security.protected(View)
    def getCandidateLocalRoles(self, obj):
        """ What local roles can I assign?
        """
        member = self.getAuthenticatedMember()
        member_roles = member.getRolesInContext(obj)
        if _checkPermission(ManageUsers, obj):
            local_roles = self.getPortalRoles()
            if 'Manager' not in member_roles:
                local_roles.remove('Manager')
        else:
            local_roles = [ role for role in member_roles
                            if role not in ('Member', 'Authenticated') ]
        return tuple(sorted(local_roles))

    @security.protected(View)
    @postonly
    def setLocalRoles(self, obj, member_ids, member_role, reindex=1,
                      REQUEST=None):
        """ Add local roles on an item.
        """
        if (_checkPermission(ChangeLocalRoles, obj)
            and member_role in self.getCandidateLocalRoles(obj)):
            for member_id in member_ids:
                roles = list(obj.get_local_roles_for_userid(userid=member_id))

                if member_role not in roles:
                    roles.append(member_role)
                    obj.manage_setLocalRoles(member_id, roles)

        if reindex and hasattr(aq_base(obj), 'reindexObjectSecurity'):
            obj.reindexObjectSecurity()

    @security.protected(View)
    @postonly
    def deleteLocalRoles(self, obj, member_ids, reindex=1, recursive=0,
                         REQUEST=None):
        """ Delete local roles of specified members.
        """
        if _checkPermission(ChangeLocalRoles, obj):
            for member_id in member_ids:
                if obj.get_local_roles_for_userid(userid=member_id):
                    obj.manage_delLocalRoles(userids=member_ids)
                    break

        if recursive and hasattr(aq_base(obj), 'contentValues'):
            for subobj in obj.contentValues():
                self.deleteLocalRoles(subobj, member_ids, 0, 1)

        if reindex and hasattr(aq_base(obj), 'reindexObjectSecurity'):
            # reindexObjectSecurity is always recursive
            obj.reindexObjectSecurity()

    @security.private
    def addMember(self, id, password, roles, domains, properties=None):
        '''Adds a new member to the user folder.  Security checks will have
        already been performed.  Called by portal_registration.
        '''
        self.acl_users._doAddUser(id, password, roles, domains)

        if properties is not None:
            member = self.getMemberById(id)
            member.setMemberProperties(properties)

    @security.protected(ManageUsers)
    @postonly
    def deleteMembers(self, member_ids, delete_memberareas=1,
                      delete_localroles=1, REQUEST=None):
        """ Delete members specified by member_ids.
        """
        # Delete members in acl_users.
        acl_users = self.acl_users
        if _checkPermission(ManageUsers, acl_users):
            if isinstance(member_ids, six.string_types):
                member_ids = (member_ids,)
            member_ids = list(member_ids)
            for member_id in member_ids[:]:
                if not acl_users.getUserById(member_id, None):
                    member_ids.remove(member_id)
            try:
                acl_users.userFolderDelUsers(member_ids)
            except (AttributeError, NotImplementedError):
                raise NotImplementedError('The underlying User Folder '
                                         'doesn\'t support deleting members.')
        else:
            raise AccessControl_Unauthorized('You need the \'Manage users\' '
                                 'permission for the underlying User Folder.')

        # Delete member data in portal_memberdata.
        mdtool = queryUtility(IMemberDataTool)
        if mdtool is not None:
            for member_id in member_ids:
                mdtool.deleteMemberData(member_id)

        # Delete members' home folders including all content items.
        if delete_memberareas:
            for member_id in member_ids:
                self.deleteMemberArea(member_id)

        # Delete members' local roles.
        if delete_localroles:
            self.deleteLocalRoles(getUtility(ISiteRoot), member_ids,
                                  reindex=1, recursive=1)

        return tuple(member_ids)

    @security.public
    def getHomeFolder(self, id=None, verifyPermission=0):
        """Returns a member's home folder object or None.
        Set verifyPermission to 1 to return None when the user
        doesn't have the View permission on the folder.
        """
        return None

    @security.public
    def getHomeUrl(self, id=None, verifyPermission=0):
        """Returns the URL to a member's home folder or None.
        Set verifyPermission to 1 to return None when the user
        doesn't have the View permission on the folder.
        """
        return None

InitializeClass(MembershipTool)
registerToolInterface('portal_membership', IMembershipTool)


@implementer(IFactory)
class HomeFolderFactoryBase(object):

    """Creates a home folder.
    """

    title = _(u'Home Folder')
    description = _(u'A home folder for portal members.')

    def __call__(self, id, title=None, *args, **kw):
        if title is None:
            title = "{0}'s Home".format(id)
        item = PortalFolder(id, title, *args, **kw)
        item.manage_setLocalRoles(id, ['Owner'])
        return item

    def getInterfaces(self):
        return implementedBy(PortalFolder)


class _BBBHomeFolderFactory(HomeFolderFactoryBase):

    """Creates a home folder.
    """

    description = _(u'Classic CMFCore home folder for portal members.')

    def __call__(self, id, title=None, *args, **kw):
        item = super(_BBBHomeFolderFactory,
                     self).__call__(id, title=title, *args, **kw)

        item.manage_permission(View,
                               ['Owner', 'Manager', 'Reviewer'], 0)
        item.manage_permission(AccessContentsInformation,
                               ['Owner', 'Manager', 'Reviewer'], 0)
        return item

BBBHomeFolderFactory = _BBBHomeFolderFactory()



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for CMFBTreeFolder module.
"""

import unittest
import Testing

from Products.CMFCore.testing import ConformsToFolder


class CMFBTreeFolderTests(ConformsToFolder, unittest.TestCase):

    def _getTargetClass(self):
        from Products.CMFCore.CMFBTreeFolder import CMFBTreeFolder

        return CMFBTreeFolder

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_empty(self):
        empty = self._makeOne('test')
        self.assertEqual(len(empty.objectIds()), 0)

# this test doesn't work with ZopeTestCase.ZopeLite.installProduct
##    def test___module_aliases__( self ):
##        from Products.BTreeFolder2.CMFBTreeFolder \
##            import CMFBTreeFolder as BBB
##
##        self.assertTrue( BBB is self._getTargetClass() )


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(CMFBTreeFolderTests),
        ))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for DiscussionTool module.
"""

import unittest
import Testing

from zope.interface.verify import verifyClass


class DiscussionToolTests(unittest.TestCase):

    def test_interfaces(self):
        from Products.CMFCore.DiscussionTool import DiscussionTool
        from Products.CMFCore.interfaces import IActionProvider
        from Products.CMFCore.interfaces import IOldstyleDiscussionTool

        verifyClass(IActionProvider, DiscussionTool)
        verifyClass(IOldstyleDiscussionTool, DiscussionTool)


class OldDiscussableTests(unittest.TestCase):

    def test_interfaces(self):
        from Products.CMFCore.DiscussionTool import OldDiscussable
        from Products.CMFCore.interfaces import IOldstyleDiscussable

        verifyClass(IOldstyleDiscussable, OldDiscussable)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(DiscussionToolTests),
        unittest.makeSuite(OldDiscussableTests),
        ))



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for FSZSQLMethod module.
"""

import unittest
from Testing import ZopeTestCase
try:
    import Products.ZSQLMethods
    ZopeTestCase.installProduct('ZSQLMethods', 1)
    HAVE_ZSQL = True
except ImportError:
    HAVE_ZSQL = False

from os.path import join

from Acquisition import aq_base
from zope.testing.cleanup import cleanUp

from Products.CMFCore.FSMetadata import FSMetadata
from Products.CMFCore.tests.base.testcase import FSDVTest
from Products.CMFCore.tests.base.testcase import SecurityTest


class FSZSQLMaker(FSDVTest):

    def _makeOne(self, id, filename):
        from Products.CMFCore.FSZSQLMethod import FSZSQLMethod
        path = join(self.skin_path_name, filename)
        metadata = FSMetadata(path)
        metadata.read()
        return FSZSQLMethod(id, path, properties=metadata.getProperties())


@unittest.skipUnless(HAVE_ZSQL, 'Products.ZSQLMethods not installed.')
class FSZSQLMethodTests(FSDVTest):

    def setUp(self):
        FSDVTest.setUp(self)
        self._registerDirectory(self)

    def test_initialization(self):
        zsql = self.ob.fake_skin.testsql
        self.assertEqual(zsql.title, 'This is a title')
        self.assertEqual(zsql.connection_id, 'testconn')
        self.assertEqual(zsql.arguments_src, 'id')
        self.assertEqual(zsql.max_rows_, 1000)
        self.assertEqual(zsql.max_cache_, 100)
        self.assertEqual(zsql.cache_time_, 10)
        self.assertEqual(zsql.class_name_, 'MyRecord')
        self.assertEqual(zsql.class_file_, 'CMFCore.TestRecord')
        self.assertEqual(zsql.connection_hook, 'MyHook')
        self.assertFalse(zsql.allow_simple_one_argument_traversal is None)


@unittest.skipUnless(HAVE_ZSQL, 'Products.ZSQLMethods not installed.')
class FSZSQLMethodCustomizationTests(SecurityTest, FSZSQLMaker):

    def setUp(self):
        FSZSQLMaker.setUp(self)
        SecurityTest.setUp(self)
        self.skins, self.custom, self.fsdir, self.fsZSQL = self._makeContext(
                                                     'testsql', 'testsql.zsql')

    def tearDown(self):
        cleanUp()
        SecurityTest.tearDown(self)
        FSZSQLMaker.tearDown(self)

    def test_customize(self):
        self.fsZSQL.manage_doCustomize(folder_path='custom')

        self.assertEqual(len(self.custom.objectIds()), 1)
        self.assertTrue('testsql' in self.custom.objectIds())

    def test_customize_alternate_root(self):
        from OFS.Folder import Folder

        self.app.other = Folder('other')

        self.fsZSQL.manage_doCustomize(folder_path='other', root=self.app)

        self.assertFalse('testsql' in self.custom.objectIds())
        self.assertTrue('testsql' in self.app.other.objectIds())

    def test_customize_fspath_as_dot(self):
        self.fsZSQL.manage_doCustomize(folder_path='.')

        self.assertFalse('testsql' in self.custom.objectIds())
        self.assertTrue('testsql' in self.skins.objectIds())

    def test_customize_manual_clone(self):
        from OFS.Folder import Folder

        clone = Folder('testsql')

        self.fsZSQL.manage_doCustomize(folder_path='custom', obj=clone)

        self.assertTrue('testsql' in self.custom.objectIds())
        self.assertTrue(aq_base(self.custom._getOb('testsql')) is clone)

    def test_customize_properties(self):
        # Make sure all properties are coming across
        self.fsZSQL.manage_doCustomize(folder_path='custom')
        zsql = self.custom.testsql

        self.assertEqual(zsql.title, 'This is a title')
        self.assertEqual(zsql.connection_id, 'testconn')
        self.assertEqual(zsql.arguments_src, 'id')
        self.assertEqual(zsql.max_rows_, 1000)
        self.assertEqual(zsql.max_cache_, 100)
        self.assertEqual(zsql.cache_time_, 10)
        self.assertEqual(zsql.class_name_, 'MyRecord')
        self.assertEqual(zsql.class_file_, 'CMFCore.TestRecord')
        self.assertEqual(zsql.connection_hook, 'MyHook')
        self.assertFalse(zsql.allow_simple_one_argument_traversal is None)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(FSZSQLMethodTests),
        unittest.makeSuite(FSZSQLMethodCustomizationTests),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Customizable Python scripts that come from the filesystem.
"""

from difflib import unified_diff
from six import get_unbound_function

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from App.special_dtml import DTMLFile
from ComputedAttribute import ComputedAttribute
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PythonScripts.PythonScript import PythonScript
from Shared.DC.Scripts.Script import Script

from Products.CMFCore.DirectoryView import registerFileExtension
from Products.CMFCore.DirectoryView import registerMetaType
from Products.CMFCore.FSObject import FSObject
from Products.CMFCore.permissions import FTPAccess
from Products.CMFCore.permissions import View
from Products.CMFCore.permissions import ViewManagementScreens
from Products.CMFCore.utils import _dtmldir

_marker = object()


class CustomizedPythonScript(PythonScript):

    """ Subclass which captures the "source" version's text.
    """

    #meta_type = 'Customized Python Script'  #(need permissions here)

    security = ClassSecurityInfo()

    def __init__(self, id, text):
        super(CustomizedPythonScript, self).__init__(id)
        self.write(text)
        self.original_source = text

    @security.protected(ViewManagementScreens)
    def getDiff(self):
        """ Return a diff of the current source with the original source.
        """
        return unified_diff(self.original_source.splitlines(),
                            self.read().splitlines(),
                            'original',
                            'modified',
                            '',
                            '',
                            lineterm="")

    security.declareProtected(ViewManagementScreens, 'manage_showDiff')
    manage_showDiff = PageTemplateFile('www/cpsDiff.pt', globals())

    manage_options = (
        PythonScript.manage_options[:1] +
        ({'label': 'Diff', 'action': 'manage_showDiff'},) +
        PythonScript.manage_options[1:])

InitializeClass(CustomizedPythonScript)


class FSPythonScript(FSObject, Script):

    """FSPythonScripts act like Python Scripts but are not directly
    modifiable from the management interface."""

    meta_type = 'Filesystem Script (Python)'
    _params = _body = ''
    _proxy_roles = ()
    _owner = None  # Unowned

    manage_options = (
        {'label': 'Customize', 'action': 'manage_main'},)

    security = ClassSecurityInfo()
    security.declareObjectProtected(View)

    security.declareProtected(ViewManagementScreens, 'manage_main')
    manage_main = DTMLFile('custpy', _dtmldir)

    security.declareProtected(View, 'index_html',)
    # Prevent the bindings from being edited TTW
    security.declarePrivate('ZBindings_edit', 'ZBindingsHTML_editForm',
                            'ZBindingsHTML_editAction')

    def _createZODBClone(self):
        """Create a ZODB (editable) equivalent of this object."""
        return CustomizedPythonScript(self.getId(), self.read())

    def _readFile(self, reparse):
        """Read the data from the filesystem.
        """
        file = open(self._filepath, 'r')
        try:
            data = file.read()
        finally:
            file.close()

        if reparse:
            self._write(data, reparse)

    def _validateProxy(self, roles=None):
        pass

    def __render_with_namespace__(self, namespace):
        '''Calls the script.'''
        self._updateFromFS()
        return Script.__render_with_namespace__(self, namespace)

    def __call__(self, *args, **kw):
        '''Calls the script.'''
        self._updateFromFS()
        return Script.__call__(self, *args, **kw)

    _exec = get_unbound_function(PythonScript._exec)

    security.declareProtected(ViewManagementScreens, 'getModTime')
    # getModTime defined in FSObject

    def ZScriptHTML_tryParams(self):
        """Parameters to test the script with."""
        param_names = []
        for name in self._params.split(','):
            name = name.strip()
            if name and name[0] != '*':
                param_names.append(name.split('=', 1)[0])
        return param_names

    @security.protected(ViewManagementScreens)
    def read(self):
        self._updateFromFS()
        return self._source

    @security.protected(ViewManagementScreens)
    def document_src(self, REQUEST=None, RESPONSE=None):
        """Return unprocessed document source."""

        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
        return self._source

    @security.protected(ViewManagementScreens)
    def PrincipiaSearchSource(self):
        "Support for searching - the document's contents are searched."
        return "%s\n%s" % (self._params, self._body)

    @security.protected(ViewManagementScreens)
    def params(self):
        return self._params

    security.declareProtected(ViewManagementScreens, 'manage_haveProxy')
    manage_haveProxy = get_unbound_function(PythonScript.manage_haveProxy)

    @security.protected(ViewManagementScreens)
    def body(self):
        return self._body

    @security.protected(ViewManagementScreens)
    def get_size(self):
        return len(self.read())

    @security.protected(FTPAccess)
    def manage_FTPget(self):
        "Get source for FTP download"
        self.REQUEST.RESPONSE.setHeader('Content-Type', 'text/plain')
        return self.read()

    def _write(self, text, compile):
        '''
        Parses the source, storing the body, params, title, bindings,
        and source in self.  If compile is set, compiles the
        function.
        '''
        ps = PythonScript(self.id)
        ps.write(text)
        if compile:
            ps._makeFunction()
            self._v_ft = ps._v_ft
            self.__code__ = ps.__code__
            self.__defaults__ = ps.__defaults__
        self._body = ps._body
        self._params = ps._params
        self.title = ps.title
        self._setupBindings(ps.getBindingAssignments().getAssignedNames())
        self._source = ps.read()  # Find out what the script sees.

    def _func_defaults(self):
        # This ensures __code__ and __defaults__ are
        # set when the code hasn't been compiled yet,
        # just in time for mapply().  Truly odd, but so is mapply(). :P
        self._updateFromFS()
        return self.__dict__.get('__defaults__', None)
    __defaults__ = ComputedAttribute(_func_defaults, 1)

    def _func_code(self):
        # See func_defaults.
        self._updateFromFS()
        return self.__dict__.get('__code__', None)
    __code__ = ComputedAttribute(_func_code, 1)

    def title(self):
        # See func_defaults.
        self._updateFromFS()
        return self.__dict__.get('title', None)
    title = ComputedAttribute(title, 1)

    def getBindingAssignments(self):
        # Override of the version in Bindings.py.
        # This version ensures that bindings get loaded on demand.
        if not hasattr(self, '_bind_names'):
            # Set a default first to avoid recursion
            self._setupBindings()
            # Now do it for real
            self._updateFromFS()
        return self._bind_names

InitializeClass(FSPythonScript)

registerFileExtension('py', FSPythonScript)
registerMetaType('Script (Python)', FSPythonScript)
